# coding: utf-8

import logging
import resource

_logger = logging.getLogger(__name__)


def increase_max_number_file():
    """increase the maximal number of file the software can open within respect of the hard limit"""
    try:
        hard_nofile = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
        resource.setrlimit(resource.RLIMIT_NOFILE, (hard_nofile, hard_nofile))
    except (ValueError, OSError):
        _logger.warning("Failed to retrieve and set the max opened files limit")
    else:
        _logger.debug("Set max opened files to %d", hard_nofile)



"""gui settings"""

from silx.gui import qt


Y_AXIS_DOWNWARD = True
"""Default axis direction Y"""

EDITING_BACKGROUND_COLOR = qt.QColor("#c3d0d6")
"""Background color used to notify the user that application waits end of edition before taking it into account"""

TAB_LABEL_PLATFORM_SETTINGS = "platform settings (for local reconstruction)"

TAB_LABEL_RECONSTRUCTION_SETTINGS = "reconstruction settings"



from __future__ import annotations

import functools
import logging

from ewoksorange.bindings import (
    OWEwoksWidgetWithTaskStack,
    OWEwoksWidgetOneThreadPerRun,
)
from ewoksorange.bindings.owwidgets import invalid_data
from orangewidget.widget import OWBaseWidget
from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess

from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing

_logger = logging.getLogger(__name__)


class _SuperviseMixIn(SuperviseProcess):
    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        self.destroyed.connect(functools.partial(ProcessManager().unregister, self))

    def setCaption(self, caption):
        self.name = caption
        try:
            ProcessManager().process_renamed(process=self)
        except Exception as e:
            _logger.warning(f"Fail to update process name. Error is {e}")

    def notify_skip(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.SKIPPED, details=details
        )

    def notify_pending(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.PENDING, details=details
        )

    def notify_succeed(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.SUCCEED, details=details
        )

    def notify_failed(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.FAILED, details=details
        )

    def notify_on_going(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.ON_GOING, details=details
        )


class SuperviseOW(OWBaseWidget, _SuperviseMixIn, openclass=True):
    """
    A basic OWWidget but registered on the process manager
    """

    want_control_area = False

    def __init__(self, parent, process_id=None):
        OWBaseWidget.__init__(self, parent, process_id=process_id)
        _SuperviseMixIn.__init__(self, process_id=process_id)

    def setCaption(self, caption):
        OWBaseWidget.setCaption(self, caption)
        _SuperviseMixIn.setCaption(self, caption=caption)


class TomwerWithStackStack(
    OWEwoksWidgetWithTaskStack, _SuperviseMixIn, WidgetLongProcessing, openclass=True
):
    def __init__(self, parent, process_id=None, *args, **kwargs):
        OWEwoksWidgetWithTaskStack.__init__(self, parent, args, kwargs)
        _SuperviseMixIn.__init__(self, process_id=process_id)
        WidgetLongProcessing.__init__(self)

        self.task_executor_queue.sigComputationStarted.connect(self._startProcessing)
        self.task_executor_queue.sigComputationEnded.connect(self._endProcessing)

    def setCaption(self, caption):
        OWBaseWidget.setCaption(self, caption)
        _SuperviseMixIn.setCaption(self, caption=caption)

    def trigger_downstream(self) -> None:
        # for now ewoksorange send ewoks variable. This will work only if
        # all task are implemented using ewokwidget which is not the case today
        for ewoksname, var in self.get_task_outputs().items():
            channel = self._get_output_signal(ewoksname)
            if invalid_data.is_invalid_data(var.value):
                channel.send(None)  # or channel.invalidate?
            else:
                channel.send(var.value)


class TomwerUnsupervisedOneThreadPerRun(
    OWEwoksWidgetOneThreadPerRun, WidgetLongProcessing, openclass=True
):
    """EwoksOrange OWEwoksWidgetOneThreadPerRun with API to provide dataset state information"""

    pass



"""Set of icons for buttons.

Use :func:`getQIcon` to create Qt QIcon from the name identifying an icon.
"""

import logging
import os
import weakref

from silx.gui import qt

import tomwer.resources

_logger = logging.getLogger(__name__)
"""Module logger"""


_cached_icons = weakref.WeakValueDictionary()
"""Cache loaded icons in a weak structure"""


_supported_formats = None
"""Order of file format extension to check"""


def getQIcon(name):
    """Create a QIcon from its name.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx".

    If no prefix are specified, the file with be returned from the silx
    resource directory with a specific path "gui/icons".

    See also :func:`silx.resources.register_resource_directory`.

    :param name: Name of the icon, in one of the defined icons
                     in this module.
    :return: Corresponding QIcon
    :raises: ValueError when name is not known
    """
    qfile = getQFile(name)
    pixmap = qt.QPixmap(qfile.fileName())
    icon = qt.QIcon(pixmap)
    return icon


def getQPixmap(name):
    """Create a QPixmap from its name.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx".

    If no prefix are specified, the file with be returned from the silx
    resource directory with a specific path "gui/icons".

    See also :func:`silx.resources.register_resource_directory`.

    :param name: Name of the icon, in one of the defined icons
                     in this module.
    :return: Corresponding QPixmap
    :raises: ValueError when name is not known
    """
    qfile = getQFile(name)
    if qfile is None:
        raise ValueError("Not an illustration name: %s" % name)
    return qt.QPixmap(qfile.fileName())


def getQFile(name) -> qt.QFile:
    """Create a QFile from an icon name. Filename is found
    according to supported Qt formats.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx".

    If no prefix are specified, the file with be returned from the silx
    resource directory with a specific path "gui/icons".

    See also :func:`silx.resources.register_resource_directory`.

    :param name: Name of the icon, in one of the defined icons
                     in this module.
    :return: Corresponding QFile
    :raises: ValueError when name is not known
    """
    # if the extension is given in the name, skip the supported_formats
    _name = name.replace(" ", "_")
    if "." in _name:
        return _getQFile(_name)
    global _supported_formats
    if _supported_formats is None:
        _supported_formats = []
        supported_formats = qt.supportedImageFormats()
        order = ["mng", "gif", "svg", "png", "jpg"]
        for format_ in order:
            if format_ in supported_formats:
                _supported_formats.append(format_)
        if len(_supported_formats) == 0:
            _logger.error("No format supported for icons")
        else:
            _logger.debug("Format %s supported", ", ".join(_supported_formats))

    for format_ in _supported_formats:
        format_ = str(format_)
        out = _getQFile("%s.%s" % (_name, format_))
        if out and out.exists():
            return out
    raise ValueError("Not an illustration name: %s" % _name)


def getResourceFileName(name):
    return tomwer.resources._resource_filename(
        name, default_directory=os.path.join("gui", "illustrations")
    )


def _getQFile(name):
    filename = tomwer.resources._resource_filename(
        name, default_directory=os.path.join("gui", "illustrations")
    )
    qfile = qt.QFile(filename)
    if qfile.exists():
        return qfile
    else:
        return None



"""
(Ewoks)Orange Widgets for debugging. To be used at your own risk and peril
"""

NAME = "debug tools"

ID = "orangecontrib.tomwer.widgets.debugtools"

DESCRIPTION = "some widgets to help debugging"

LONG_DESCRIPTION = "some widgets to help debugging"

ICON = "../../widgets/icons/tomwer_debug_tools.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 20



# coding: utf-8
from __future__ import annotations


import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess

from silx.gui import qt
from silx.gui.utils import blockSignals

from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.process.reconstruction.nabu.nabuslices import NabuSlicesTask
from tomwer.core.process.reconstruction.nabu.nabuvolume import NabuVolumeTask

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class NabuSliceProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisTask` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, parent=None, process_id=None):
        qt.QObject.__init__(self, parent=parent)
        FIFO.__init__(self, process_id=process_id)
        self._dry_run = False

    def _process(self, data, configuration, callback=None):
        _logger.info(f"Nabu slice stack is processing {data}")
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )

        self._data_currently_computed = data
        assert isinstance(data, TomwerScanBase)
        self._computationThread.finished.connect(self._end_threaded_computation)

        if isOnLbsram(data) and is_low_on_memory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = f"low memory, skip nabu reconstruction for {data.path}"
            _logger.processSkipped(mess)
            ProcessManager().notify_dataset_state(
                dataset=data,
                process=self,
                state=DatasetState.SKIPPED,
            )
            self._end_threaded_computation()
        else:
            self._computationThread.init(data=data, configuration=configuration)
            self._computationThread.setDryRun(self._dry_run)
            # need to manage connect before starting it because
            self._computationThread.start()

    def _end_threaded_computation(self, callback=None):
        self._computationThread.finished.disconnect(self._end_threaded_computation)
        super()._end_threaded_computation(callback=callback)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _SliceProcessingThread(process_id=process_id)

    def setDryRun(self, dry_run):
        self._dry_run = dry_run

    def cancel(self):
        if self._computationThread.isRunning():
            with blockSignals(self._computationThread):
                self._computationThread.cancel()
                # stop stack
                super().stop()
                # emit next
                self.sigComputationEnded.emit(None, None)


class NabuVolumeProcessStack(NabuSliceProcessStack):
    """Implementation of the `.AxisTask` but having a stack for treating
    scans and making computation in threads"""

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _VolumeProcessingThread(process_id=process_id)


class _SliceProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of nabu reconstruction
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        self._scan = None
        self._future_tomo_obj = None
        self._configuration = None
        self._dry_run = False
        self._current_processing = None
        # processing currently runned. As we want to be able to cancel processing we need to keep a pointer on the
        # 'activate' nabu processing to stop process if requested
        self._task = None
        # ewoks task for reconstruction

    @property
    def future_tomo_obj(self):
        return self._future_tomo_obj

    def setDryRun(self, dry_run):
        self._dry_run = dry_run

    def init(self, data, configuration):
        self._scan = data
        self._configuration = configuration
        self._task = None

    def run(self):
        # note: now rnu does a processing close to the 'run_slices_reconstruction' except that we keep a trace on the
        # current nabu subprocess runned. The goal is to be able to cancel / stop processing when asked
        self.sigComputationStarted.emit()
        mess = f"Start nabu slice(s) reconstruction of {self._scan}"
        _logger.processStarted(mess)
        ProcessManager().notify_dataset_state(
            dataset=self._scan,
            process=self,
            state=DatasetState.ON_GOING,
            details=mess,
        )

        self._task = NabuSlicesTask(
            process_id=self.process_id,
            inputs={
                "data": self._scan,
                "nabu_params": self._configuration,
                "dry_run": self._dry_run,
                "serialize_output_data": False,
                "invalid_slice_callback": self.showErrorSliceInvalid,
            },
        )
        self._task.run()
        self._future_tomo_obj = self._task.outputs.future_tomo_obj

    def cancel(self):
        if self._task is not None:
            self._task.cancel()
        self.quit()

    def showErrorSliceInvalid(self, error_message):
        self.sigError.emit(error_message)


class _VolumeProcessingThread(_SliceProcessingThread):
    """
    Thread use to execute the processing of nabu reconstruction
    """

    def run(self):
        self.sigComputationStarted.emit()
        mess = f"Start nabu volume reconstruction of {self._scan}"
        _logger.processStarted(mess)
        ProcessManager().notify_dataset_state(
            dataset=self._scan,
            process=self,
            state=DatasetState.ON_GOING,
            details=mess,
        )

        self._task = NabuVolumeTask(
            process_id=self.process_id,
            inputs={
                "data": self._scan,
                "nabu_params": self._scan.nabu_recons_params,
                "nabu_extra_params": self._configuration,
                "process_id": self.process_id,
                "dry_run": self._dry_run,
                "serialize_output_data": False,  # avoid spending time on serialization / deserialization
            },
        )
        self._task.run()
        self._future_tomo_obj = self._task.outputs.future_tomo_obj



# coding: utf-8
from __future__ import annotations


import tempfile

import pytest
from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt
from nxtomo.nxobject.nxdetector import ImageKey

from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.edit.imagekeyeditor import (
    ImageKeyDialog,
    ImageKeyUpgraderWidget,
    _AddImageKeyUpgradeOperation,
)
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestImageKeyEditorGUI(TestCaseQt):
    """
    Simple test the interface for configuration is working
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = ImageKeyDialog(parent=None)
        self.output_folder = tempfile.mkdtemp()

        hdf5_mock = MockNXtomo(
            scan_path=self.output_folder,
            n_ini_proj=20,
            n_proj=20,
        )
        self._scan = hdf5_mock.scan

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None

    def testEdition(self):
        modifications = {
            2: ImageKey.INVALID,
        }
        self._widget.setScan(self._scan)
        self._widget.setModifications(modifications)
        self.assertEqual(self._widget.getModifications(), modifications)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class Test_AddImageKeyUpgradeOperation(TestCaseQt):
    def setUp(self):
        super().setUp()
        self._widget = _AddImageKeyUpgradeOperation()
        self._signalLisener = SignalListener()
        self._widget.sigOperationAdded.connect(self._signalLisener)

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        super().tearDown()

    def test(self):
        self._widget._addButtonPB.released.emit()
        self.qapp.processEvents()
        assert self._signalLisener.callCount() == 1
        assert self._signalLisener.callCount() == 1
        assert self._signalLisener.arguments() == [
            (
                {
                    "from_image_key": ImageKey.PROJECTION.value,
                    "to_image_key": ImageKey.DARK_FIELD.value,
                },
            ),
        ]


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestImageKeyUpgraderList(TestCaseQt):
    def setUp(self):
        super().setUp()
        self._widget = ImageKeyUpgraderWidget()

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        super().tearDown()

    def test_add_remove_operations(self):
        """
        test adding and remove operations to the list
        """
        self._widget.addOperation(
            from_image_key=ImageKey.PROJECTION, to_image_key=ImageKey.DARK_FIELD
        )
        assert self._widget.getOperations() == {
            ImageKey.PROJECTION: ImageKey.DARK_FIELD,
        }
        self._widget.addOperation(
            from_image_key=ImageKey.FLAT_FIELD, to_image_key=ImageKey.DARK_FIELD
        )
        assert self._widget.getOperations() == {
            ImageKey.PROJECTION: ImageKey.DARK_FIELD,
            ImageKey.FLAT_FIELD: ImageKey.DARK_FIELD,
        }
        self._widget.addOperation(
            from_image_key=ImageKey.PROJECTION, to_image_key=ImageKey.FLAT_FIELD
        )
        assert self._widget.getOperations() == {
            ImageKey.PROJECTION: ImageKey.FLAT_FIELD,
            ImageKey.FLAT_FIELD: ImageKey.DARK_FIELD,
        }
        self._widget.removeOperation(
            from_image_key=ImageKey.PROJECTION, to_image_key=ImageKey.FLAT_FIELD
        )
        assert self._widget.getOperations() == {
            ImageKey.FLAT_FIELD: ImageKey.DARK_FIELD,
        }
        self._widget.removeOperation(
            from_image_key=ImageKey.PROJECTION, to_image_key=ImageKey.DARK_FIELD
        )
        assert self._widget.getOperations() == {
            ImageKey.FLAT_FIELD: ImageKey.DARK_FIELD,
        }
        self._widget.removeOperation(
            from_image_key=ImageKey.FLAT_FIELD, to_image_key=ImageKey.DARK_FIELD
        )
        assert self._widget.getOperations() == {}

        self._widget.setOperations(
            {
                ImageKey.PROJECTION: ImageKey.FLAT_FIELD,
                ImageKey.FLAT_FIELD: ImageKey.DARK_FIELD,
            }
        )
        assert self._widget.getOperations() == {
            ImageKey.PROJECTION: ImageKey.FLAT_FIELD,
            ImageKey.FLAT_FIELD: ImageKey.DARK_FIELD,
        }



# coding: utf-8
from __future__ import annotations


import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.settings import SlurmSettings
from tomwer.gui.cluster.slurm import (
    is_op_account,
    SlurmSettingsMode,
    SlurmSettingsWidget,
    SlurmSettingsWindow,
    _WallTime,
)
from tomwer.tests.utils import skip_gui_test
from tomwer.tests.conftest import qtapp  # noqa F401


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSlurmWidget(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)
        self.slurmWidget = SlurmSettingsWidget(parent=None)

    def tearDown(self):
        self.slurmWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.slurmWidget.close()
        self.slurmWidget = None

    def testGetConfiguration(self):
        dict_res = self.slurmWidget.getConfiguration()
        expected_dict = {
            "cpu-per-task": SlurmSettings.N_CORES_PER_TASK,
            "n_tasks": SlurmSettings.N_TASKS,
            "memory": SlurmSettings.MEMORY_PER_WORKER,
            "partition": "",
            "n_gpus": SlurmSettings.N_GPUS_PER_WORKER,
            "job_name": "tomwer_{scan}_-_{process}_-_{info}",
            "n_jobs": 1,
            "modules": ("tomotools/stable",),
            "walltime": "01:00:00",
            "sbatch_extra_params": {
                "export": "ALL",
                "gpu_card": None,
            },
        }
        assert dict_res == expected_dict, f"{dict_res} vs {expected_dict}"

    def testSetConfiguration(self):
        self.slurmWidget.setConfiguration(
            {
                "cpu-per-task": 2,
                "n_tasks": 3,
                "memory": 156,
                "partition": "test-queue",
                "n_gpus": 5,
                "modules": "mymodule, mysecond/10.3",
                "sbatch_extra_params": {
                    "export": "NONE",
                },
            }
        )

        assert self.slurmWidget.getNCores() == 2
        assert self.slurmWidget.getNWorkers() == 3
        assert self.slurmWidget.getMemory() == 156
        assert self.slurmWidget.getQueue() == "test-queue"
        assert self.slurmWidget.getNGPU() == 5
        assert self.slurmWidget.getSBatchExtraParams() == {
            "export": "NONE",
            "gpu_card": None,
        }


def test_SlurmSettingsWindow(qtapp):  # noqa F811
    """
    test that slurm SettingsWindow can load all the different slurm settings modes
    """
    widget = SlurmSettingsWindow()
    for mode in SlurmSettingsMode:
        widget.setCurrentSettingsMode(mode)
        widget.getConfiguration()


def test_is_op_account():
    assert is_op_account("opid19")
    assert not is_op_account("opid0")
    assert not is_op_account("dopid05")
    assert not is_op_account("opid")
    assert is_op_account("opd12")


def test__WallTime():
    assert _WallTime("11:00:02") > _WallTime("10:25:45")
    assert _WallTime("1-00:00:00") > _WallTime("10:25:45")
    assert _WallTime("01:00:0") < _WallTime("01:00:01")
    assert _WallTime("1-00:00:00") == _WallTime(
        "24:00:0"
    ), f"{_WallTime('1-00:00:00').value} != {_WallTime('24:00:0').value} "



# coding: utf-8
from __future__ import annotations

import pathlib
from datetime import datetime
from processview.core.dataset import Dataset, DatasetIdentifier
from tomoscan.esrf.identifier.edfidentifier import (
    EDFVolumeIdentifier as _EDFVolumeIdentifier,
)
from tomoscan.esrf.volume.edfvolume import EDFVolume as _EDFVolume

from tomwer.core.volume.volumebase import TomwerVolumeBase


class EDFVolumeIdentifier(_EDFVolumeIdentifier, DatasetIdentifier):
    def __init__(self, object, folder, file_prefix, metadata=None):
        super().__init__(object, folder, file_prefix)
        DatasetIdentifier.__init__(
            self, data_builder=EDFVolume.from_identifier, metadata=metadata
        )

    @staticmethod
    def from_str(identifier):
        return _EDFVolumeIdentifier._from_str_to_single_frame_identifier(
            identifier=identifier,
            SingleFrameIdentifierClass=EDFVolumeIdentifier,
            ObjClass=EDFVolume,
        )

    def long_description(self) -> str:
        """used for processview header tooltip for now"""
        return self.to_str()


class EDFVolume(_EDFVolume, TomwerVolumeBase, Dataset):
    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, EDFVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {EDFVolumeIdentifier} and not {type(identifier)}"
            )
        return EDFVolume(
            folder=identifier.folder,
            volume_basename=identifier.file_prefix,
        )

    def get_identifier(self) -> EDFVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")
        try:
            stat = pathlib.Path(self.url.file_path()).stat()
        except Exception:
            stat = None

        return EDFVolumeIdentifier(
            object=self,
            folder=self.url.file_path(),
            file_prefix=self._volume_basename,
            metadata={
                "name": self.url.file_path(),
                "creation_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
                "modification_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
            },
        )



import numpy
from copy import deepcopy
from ewokscore.task import Task as EwoksTask
from tomwer.model.volume_viewer import VolumeViewerModel
from tomwer.core.volume.volumebase import TomwerVolumeBase


class VolumeViewerTask(
    EwoksTask,
    input_model=VolumeViewerModel,
    output_names=("slices", "volume_metadata", "volume_shape", "loaded_volume"),
):
    """Function loading a three slices into each direction and volume metadata and create a 'summary' of the reconstructed volume"""

    N_SLICES_PER_AXIS = 3

    def run(self):
        volume = self.inputs.volume
        assert isinstance(
            volume, TomwerVolumeBase
        ), f"volume should be an instance of {TomwerVolumeBase}. got {type(volume)}"
        if self.inputs.load_volume:
            loaded_volume = deepcopy(volume)
            loaded_volume.load_data(store=True)
            self.outputs.loaded_volume = loaded_volume
        else:
            self.outputs.loaded_volume = None
        self.outputs.slices = volume.get_slices(slices=self.get_slices_to_extract())
        self.outputs.volume_metadata = (
            self.inputs.volume.metadata or self.inputs.volume.load_metadata()
        )
        self.outputs.volume_shape = volume.get_volume_shape()

    def get_slices_to_extract(self) -> tuple[tuple[int, tuple[int]]]:
        """
        Return the indices of the slices to extract on each axis (N_SLICES_PER_AXIS on each axis).

        return tuple (A) is a two elements tuple. First element if the axis (B).
        Second is the tuple of indices to extract along the axis (B)
        indices are equally spaced in each dimensions
        """
        result: list[tuple[int, tuple[int]]] = []
        volume = self.inputs.volume
        if not isinstance(volume, TomwerVolumeBase):
            raise TypeError(
                f"Volume is expected to be an instance of {TomwerVolumeBase}. Got {type(volume)}"
            )
        volume_shape = volume.get_volume_shape()
        for axis, axis_len in enumerate(volume_shape):
            for slice_index in numpy.linspace(
                0, axis_len, endpoint=False, num=self.N_SLICES_PER_AXIS + 1
            )[1:]:
                result.append((axis, numpy.round(slice_index).astype(numpy.uint16)))

        return tuple(result)



import gc
import os
import logging

import pytest
from ewoksorange.bindings.qtapp import get_all_qtwidgets, qtapp_context

_logger = logging.getLogger(__name__)


def global_cleanup_orange():
    from orangecanvas.document.suggestions import Suggestions

    Suggestions.instance = None


def global_cleanup_pytest():
    for obj in gc.get_objects():
        if isinstance(obj, logging.LogRecord):
            obj.exc_info = None  # traceback keeps frames which keep locals


def collect_garbage(app):
    if app is None:
        return
    app.processEvents()
    while gc.collect():
        app.processEvents()


@pytest.fixture(scope="session")
def qtapp():
    """
    create a Qt application if doesn't exists
    """
    with qtapp_context() as app:
        yield app
    collect_garbage(app)
    global_cleanup_orange()
    global_cleanup_pytest()
    collect_garbage(app)
    warn_qtwidgets_alive()


def warn_qtwidgets_alive():
    widgets = get_all_qtwidgets()
    if widgets:
        _logger.warning(
            "%d remaining widgets after tests:\n %s",
            len(widgets),
            "\n ".join(map(str, widgets)),
        )


@pytest.fixture(scope="function")
def nxtomo_scan_180(tmp_path):
    test_path = tmp_path / "test"
    test_path.mkdir()
    yield create_default_scan(
        scan_path=os.path.join(test_path, "scan_180"),
        scan_range=180,
    )


@pytest.fixture(scope="function")
def nxtomo_scan_360(tmp_path):
    test_path = tmp_path / "test"
    test_path.mkdir()
    yield create_default_scan(
        scan_path=os.path.join(test_path, "scan_360"),
        scan_range=360,
    )


def create_default_scan(scan_path: str, scan_range: int):
    from tomwer.core.utils.scanutils import MockNXtomo

    return MockNXtomo(
        scan_path=scan_path,
        n_proj=10,
        n_ini_proj=10,
        scan_range=scan_range,
        dim=20,
        energy=12.3,
    ).scan



# coding: utf-8
from __future__ import annotations


import shutil
import tempfile

import pytest
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.utils.scanutils import MockEDF
from tomwer.gui.control.scanselectorwidget import ScanSelectorWidget
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestScanSelector(TestCaseQt):
    """
    Simple test for the ScanSelectorWidget
    """

    def setUp(self):
        self._folder1 = tempfile.mkdtemp()
        self._folder2 = tempfile.mkdtemp()
        self._folder3 = tempfile.mkdtemp()
        for _folder in (self._folder1, self._folder2, self._folder3):
            MockEDF.mockScan(scanID=_folder, nRadio=5, nRecons=5, nPagRecons=0, dim=10)

        self.widget = ScanSelectorWidget(parent=None)

    def tearDown(self):
        shutil.rmtree(self._folder1)
        shutil.rmtree(self._folder2)
        shutil.rmtree(self._folder3)

    def test(self):
        self.widget.add(self._folder1)
        self.widget.add(self._folder2)
        self.widget.add(self._folder3)
        self.assertEqual(self.widget.dataList.n_data(), 3)
        self.widget.remove(self._folder3)
        self.assertEqual(self.widget.dataList.n_data(), 2)



# coding: utf-8
from __future__ import annotations


import os
import pint
import shutil
import tempfile

import h5py
import numpy
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.io.url import DataUrl

from tomwer.core.process.reconstruction.saaxis.params import (
    ReconstructionMode,
    SAAxisParams,
)
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.process.reconstruction.scores.params import ScoreMethod
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.reconstruction.saaxis.dimensionwidget import DimensionWidget
from tomwer.gui.reconstruction.saaxis.saaxis import SAAxisWindow
from tomwer.tests.utils import skip_gui_test


_ureg = pint.get_application_registry()


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestDimensionWidget(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        self._window = DimensionWidget(title="test window")

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()

    def testDimension(self):
        """Insure setting dimensions values and unit are correct"""
        # check initial values
        self.assertEqual(
            self._window.getQuantities(),
            (1.0 * _ureg.millimeter, 1.0 * _ureg.millimeter, 1.0 * _ureg.millimeter),
        )
        # set dim 0 to 20 mm
        self._window._dim0ValueQLE.setValue(20)
        self._window._dim0ValueQLE.editingFinished.emit()
        self.qapp.processEvents()
        self.assertEqual(
            self._window.getQuantities(cast_unit_to=_ureg.meter),
            (0.02 * _ureg.meter, 0.001 * _ureg.meter, 0.001 * _ureg.meter),
        )
        # set dim 1 to 0.6 millimeter
        self._window._dim1ValueQLE.setValue(0.6)
        self._window._dim1ValueQLE.editingFinished.emit()
        self.qapp.processEvents()
        self.assertEqual(
            self._window.getQuantities(cast_unit_to=_ureg.meter),
            (0.02 * _ureg.meter, 0.0006 * _ureg.meter, 0.001 * _ureg.meter),
        )
        # change display to micrometer
        self._window.setUnit(_ureg.millimeter)
        self.assertEqual(
            self._window.getQuantities(cast_unit_to=_ureg.meter),
            (0.02 * _ureg.meter, 0.0006 * _ureg.meter, 0.001 * _ureg.meter),
        )
        self.assertEqual(
            self._window._dim0Value,
            20 * _ureg.millimeter,
        )
        self._window._dim2ValueQLE.setValue(500)
        self._window._dim2ValueQLE.editingFinished.emit()
        self.qapp.processEvents()
        self.assertEqual(
            self._window.getQuantities(cast_unit_to=_ureg.meter),
            (0.02 * _ureg.meter, 0.0006 * _ureg.meter, 0.5 * _ureg.meter),
        )

    def testConstructorWithColors(self):
        """Insure passing colors works"""
        DimensionWidget(title="title", dims_colors=("#ffff5a", "#62efff", "#ff5bff"))

    def testConstructorWithDimNames(self):
        """Insure passing colors works"""
        DimensionWidget(title="title", dims_name=("x", "y", "z"))


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSAAxisWindow(TestCaseQt):
    """Test the SAAxisWindow interface"""

    _N_OUTPUT_URLS = 10

    def setUp(self):
        super().setUp()
        self._window = SAAxisWindow()
        self.folder = tempfile.mkdtemp()
        self._output_urls = []
        self._cor_scores = {}

        for i in range(self._N_OUTPUT_URLS):
            output_file = os.path.join(self.folder, f"recons_{i}.h5")
            with h5py.File(output_file, mode="a") as h5f:
                h5f["data"] = numpy.random.random(100 * 100).reshape(100, 100)
                url = DataUrl(file_path=output_file, data_path="data", scheme="silx")
                assert url.is_valid()
            self._output_urls.append(url)
            score = ComputedScore(
                tv=numpy.random.randint(10),
                std=numpy.random.randint(100),
            )
            self._cor_scores[i] = (url, score)

        # create a scan
        self.folder = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=self.folder, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
        )
        mock.add_alignment_radio(index=10, angle=90)
        mock.add_alignment_radio(index=10, angle=0)
        self.scan = mock.scan

        self._window.setScan(self.scan)

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        shutil.rmtree(self.folder)
        super().tearDown()

    def testGetConfiguration(self):
        """Insure configuration getter works"""
        conf = self._window.getConfiguration()
        # the configuration should be convertible to
        conf = SAAxisParams.from_dict(conf)
        self.assertEqual(conf.slice_indexes, "middle")
        self.assertEqual(conf.file_format, "hdf5")
        self.assertTrue(isinstance(conf.nabu_recons_params, dict))
        for key in (
            "preproc",
            "reconstruction",
            "dataset",
            "tomwer_slices",
            "output",
            "phase",
        ):
            with self.subTest(key=key):
                self.assertTrue(key in conf.nabu_recons_params)
        self.assertEqual(conf._dry_run, False)
        self.assertEqual(conf.mode, ReconstructionMode.VERTICAL)
        self.assertEqual(conf.score_method, ScoreMethod.TV_INVERSE)
        self.assertEqual(conf.output_dir, None)
        self.assertEqual(conf.scores, None)

    def testSetConfiguration(self):
        """Insure configuration setter works"""
        self._window.setSlicesRange(0, 20)
        configuration = {
            "slice_index": {"Slice": 2},
            "estimated_cor": 0.24,
            "research_width": 0.8,
            "n_reconstruction": 5,
        }
        self._window.setConfiguration(configuration)
        res_conf = self._window.getConfiguration()
        for key in (
            "slice_index",
            "estimated_cor",
            "research_width",
            "n_reconstruction",
        ):
            self.assertEqual(configuration[key], res_conf[key])

    def testSetResults(self):
        """Test setting results and saving result to a folder"""
        self._window.setCorScores(self._cor_scores, score_method="standard deviation")
        # test saving snapshots
        with tempfile.TemporaryDirectory() as output_png_imgs:
            final_dir = os.path.join(output_png_imgs, "test/create/it")
            self._window.saveReconstructedSlicesTo(final_dir)
            assert os.path.exists(final_dir)
            assert len(os.listdir(final_dir)) == len(self._cor_scores)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "21/06/2021"


import gc
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.debugtools.DatasetGeneratorOW import (
    DatasetGeneratorOW,
)


class TestDatasetGeneratorOW(TestCaseQt):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.widget = DatasetGeneratorOW()
        self.widget.generatorGUI.mainWidget.setRootDir(self.tmp_dir)

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        gc.collect()

    def test(self):
        self.widget._generate().join()



# coding: utf-8
"""
Utils for slurm
"""

from __future__ import annotations

import uuid
from datetime import datetime

from sluurp.utils import has_sbatch_available


def is_slurm_available() -> bool:
    """
    Return True if the environment knows about slurm command (sbatch)
    """
    return has_sbatch_available()


def get_slurm_script_name(prefix: str | None = None) -> str:
    now_str = datetime.now().strftime("%y-%m-%d_%H:%M:%S")
    random_str = str(uuid.uuid4()).split("-")[0]
    suffix = f"{now_str}_{random_str}_.sh"
    if prefix in ("", None):
        return suffix
    else:
        return f"{prefix}_{suffix}"



# coding: utf-8

from tomwer.core.process.visualization.imagestackviewer import (
    _ImageStackViewerPlaceHolder,
)


def test_image_stack_viewer():
    process = _ImageStackViewerPlaceHolder(
        inputs={
            "data": None,
        }
    )
    process.run()



from __future__ import annotations

import weakref
from silx.gui import qt

from silx.gui import icons as silx_icons


class ColorBarAction(qt.QAction):
    """
    QAction opening the ColorBarWidget of the specified plot.
    It was not possible to inherit the one from silx because it inherit from a PlotAction which is not the case of this one (n plots...)
    """

    def __init__(self, colorbar, parent=None):
        icon = silx_icons.getQIcon("colorbar")
        text = "Colorbar"
        super().__init__(parent=parent, icon=icon, text=text)

        self.setToolTip("Show/Hide the colorbar")
        self.setCheckable(True)

        self.triggered[bool].connect(self._actionTriggered)
        colorbar.sigVisibleChanged.connect(self._widgetVisibleChanged)
        self._colorBar = weakref.ref(colorbar)

    def _widgetVisibleChanged(self, isVisible):
        """Callback when the colorbar `visible` property change."""
        if self.isChecked() == isVisible:
            return
        self.setChecked(isVisible)

    def _actionTriggered(self, checked=False):
        """Create a cmap dialog and update active image and default colormap."""

        colorBar = self._colorBar()
        if colorBar is None:
            return
        if not colorBar.isHidden() == checked:
            return

        colorBar.setVisible(checked)



from ewokscore.task import Task as EwoksTask


class _TomoobjseriesPlaceHolder(
    EwoksTask, input_names=["series"], output_names=["series"]
):
    """
    task to define a tomography 'serie'
    """

    def run(self):
        self.outputs.s = self.inputs.series



"""contain the SADeltaBetaProcess. Half automatic best delta / beta finder"""

from __future__ import annotations

import logging
import os
from copy import copy, deepcopy

import h5py
import numpy
from multiprocessing import Pool
from tqdm import tqdm

from nabu.pipeline.config import get_default_nabu_config
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from tomoscan.esrf.scan.utils import get_data
from tomoscan.io import HDF5File
from silx.io.utils import h5py_read_dataset

import tomwer.version
from tomwer.core.process.reconstruction.nabu.nabucommon import (
    ResultsLocalRun,
    ResultSlurmRun,
    ResultsWithStd,
)
from tomwer.core.process.reconstruction.nabu.nabuslices import SingleSliceRunner
from tomwer.core.process.reconstruction.utils.cor import relative_pos_to_absolute
from tomwer.core.process.reconstruction.scores import (
    ComputedScore,
    ScoreMethod,
    apply_roi,
    compute_score,
    get_disk_mask_radius,
)
from tomwer.core.process.task import Task
from tomwer.core.process import utils as core_utils
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils import logconfig
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.io.utils import format_stderr_stdout
from tomwer.io.utils.h5pyutils import EntryReader
from tomwer.core.process.reconstruction.nabu.nabuscores import (
    run_nabu_one_slice_several_config,
)
from tomwer.core.futureobject import FutureTomwerObject

from silx.io.url import DataUrl
from silx.utils.deprecation import deprecated, deprecated_warning

from ..nabu import utils as nabu_utils
from .params import SADeltaBetaParams

_logger = logging.getLogger(__name__)


DEFAULT_RECONS_FOLDER = "multi_delta_beta_results"


def one_slice_several_db(
    scan: TomwerScanBase,
    configuration: dict | SADeltaBetaParams,
    process_id: int | None = None,
) -> tuple:
    """
    Run a slice reconstruction using nabu per Center Of Rotation (cor) provided
    Then for each compute a score (quality) of the center of rotation

    :param scan:
    :param configuration:
    :return: cor_reconstructions, outs, errs
             cor_reconstructions is a dictionary of cor as key and a tuple
             (url, score) as value
    """
    if isinstance(configuration, SADeltaBetaParams):
        configuration = configuration.to_dict()

    task = SADeltaBetaTask(
        process_id=process_id,
        inputs={
            "data": scan,
            "sa_axis_params": configuration,
            "serialize_output_data": False,
        },
    )
    task.run()
    return (
        task.outputs.scores,
        task.outputs.std_out,
        task.outputs.std_err,
        task.outputs.rois,
    )


class SADeltaBetaTask(
    Task,
    SuperviseProcess,
    input_names=("data", "sa_delta_beta_params"),
    output_names=("data", "best_db"),
    optional_input_names=(
        "dry_run",
        "dump_roi",
        "dump_process",
        "serialize_output_data",
        "compute_scores",  # for GUI we want to post pone the score calculation
        "pool_size",
    ),
):
    """
    Main process to launch several reconstruction of a single slice with
    several Center Of Rotation (cor) values
    """

    DEFAULT_POOL_SIZE = 10

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = inputs.get("dry_run", False)
        self._dump_process = inputs.get("dump_process", True)
        self._dump_roi = inputs.get("dump_roi", False)
        self._sa_delta_beta_params = inputs.get("sa_delta_beta_params", None)
        self._std_outs = tuple()
        self._std_errs = tuple()
        self._cancelled = False

    @property
    def dump_roi(self):
        return self._dump_roi

    @dump_roi.setter
    def dump_roi(self, dump):
        self._dump_roi = dump

    @property
    def std_outs(self):
        return self._std_outs

    @property
    def std_errs(self):
        return self._std_errs

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @deprecated(replacement="ewoks task inputs.sa_delta_beta_params")
    def set_configuration(self, configuration: dict) -> None:
        if isinstance(configuration, SADeltaBetaParams):
            self._settings = configuration.to_dict()
        elif isinstance(configuration, dict):
            self._settings = configuration
        else:
            raise TypeError(
                "configuration should be an instance of dict or " "SAAxisParams"
            )

    @staticmethod
    def autofocus(scan) -> float | None:
        scores = scan.sa_delta_beta_params.scores
        if scores is None:
            return
        score_method = scan.sa_delta_beta_params.score_method
        best_db, best_score = None, 0
        for cor, (_, score_cls) in scores.items():
            if score_cls is None:  # if score calculation failed
                continue
            score = score_cls.get(score_method)
            if score is None:
                continue
            if score > best_score:
                best_db, best_score = cor, score
        scan.sa_delta_beta_params.autofocus = best_db
        scan.sa_delta_beta_params.value = best_db
        return best_db

    def get_output_dir(self, params: SADeltaBetaParams, scan: TomwerScanBase):
        output_dir = params.output_dir or None

        if params.output_dir is None:
            output_dir = (
                params.nabu_recons_params.get("output", {}).get("location", None)
                or None
            )
            if output_dir is None:
                output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)
        return output_dir

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            self.outputs.data = scan
            return
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(f"input type of {scan}: {type(scan)} is not managed")

        config = copy(self.inputs.sa_delta_beta_params)
        axis = config.get("reconstruction", {}).get("slice_plane", "XY")
        params = SADeltaBetaParams.from_dict(config)

        # insure scan contains some parameter regarding sa delta / beta
        if scan.sa_delta_beta_params is None:
            scan.sa_delta_beta_params = params

        # insure it also contains some axis_params
        if scan.axis_params is None:
            from tomwer.core.process.reconstruction.axis import AxisRP

            scan.axis_params = AxisRP()

        # create dir if does not exists
        params.output_dir = self.get_output_dir(params=params, scan=scan)
        if not os.path.exists(params.output_dir):
            os.makedirs(params.output_dir)

        slice_index = self._preprocess_slice_index(params.slice_indexes)
        delta_beta_s = params.delta_beta_values
        # TODO: check: dry run should only be settable at one location
        dry_run = self._dry_run or params.dry_run
        cluster_config = params.cluster_config
        nabu_config = params.nabu_recons_params

        # step one: complete nabu configuration(s)
        configs = self._config_preprocessing(
            scan=scan,
            config=nabu_config,
            delta_beta_s=delta_beta_s,
            output_dir=params.output_dir,
        )
        # step 2: run reconstructions
        advancement = tqdm(
            desc=f"sa-delta-beta - slice {slice_index} of {scan.get_identifier().short_description()}"
        )

        dbs_res = {}
        rois = {}

        try:
            (
                _,
                dbs_res,
                future_tomo_objs,
                self._std_outs,
                self._std_errs,
            ) = self._run_one_slice_several_db(
                scan=scan,
                configs=configs,
                advancement=advancement,
                slice_index=slice_index,
                dry_run=dry_run,
                cluster_config=cluster_config,
                axis=axis,
            )
        except Exception as e:
            _logger.error(e)
            mess = f"sa-delta-beta -nabu- computation for {str(scan)} failed."
            state = DatasetState.FAILED
        else:
            # step 3: wait for future if any
            self._resolve_futures(
                scan=scan,
                nabu_config=next(iter(configs.items()))[
                    1
                ],  # db is not used but paganin and other parameters are. Take the first nabu configuration available
                slice_index=slice_index,
                db_reconstructions=dbs_res,
                future_tomo_objs=future_tomo_objs,
                axis=axis,
            )

            # step 4: run post processing (compute score for each slice)
            if self.get_input_value("compute_scores", True):
                try:
                    dbs_res, rois = self._post_processing(
                        scan=scan,
                        db_reconstructions=dbs_res,
                    )
                except Exception as e:
                    _logger.error(e)
                    mess = f"sa-delta-beta -post-processing- computation for {str(scan)} failed."
                    state = DatasetState.FAILED
                    dbs_res = {}
                else:
                    state = DatasetState.WAIT_USER_VALIDATION
                    self.delta_beta_s = scan.sa_delta_beta_params.autofocus
                    mess = "sa-delta-beta computation succeeded"
            else:
                dbs_res = {}
                state = DatasetState.FAILED
                mess = "couldn't find 'compute_scores'"

        if self._cancelled:
            state = DatasetState.CANCELLED
            mess = "scan cancelled by the user"

        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=state,
            details=self._compute_mess_details(mess),
        )

        scan.sa_delta_beta_params.scores = dbs_res
        best_db = self.autofocus(scan=scan)
        # store nabu recons parameters to be used within the nabu volume for example.

        sc_config = get_default_nabu_config(nabu_fullfield_default_config)
        sc_config.update(nabu_config)
        if best_db is not None:
            sc_config["phase"]["delta_beta"] = (
                best_db,
            )  # warning: at this tage delta_beta expects a list of value
        # store used reconstruction parameters - to be used later on
        scan.nabu_recons_params = sc_config

        # end processing
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan
        self.outputs.best_db = best_db

        self._process_end(scan=scan, db_res=dbs_res, score_rois=rois)

    def _compute_mess_details(self, mess=""):
        """
        util to join a message and nabu std err and std out
        """
        nabu_logs = []
        for std_err, std_out in zip(self._std_errs, self.std_outs):
            nabu_logs.append(format_stderr_stdout(stdout=std_out, stderr=std_err))
        self._nabu_log = nabu_logs
        nabu_logs.insert(0, mess)
        return "\n".join(nabu_logs)

    def _config_preprocessing(self, scan, config, delta_beta_s, output_dir) -> dict:
        config.get("phase", {}).pop("beam_shape", None)
        if (
            scan.axis_params is not None
            and scan.axis_params.relative_cor_value is not None
        ):
            if "reconstruction" in config:
                cor_nabu_ref = relative_pos_to_absolute(
                    relative_pos=scan.axis_params.relative_cor_value,
                    det_width=scan.dim_1,
                )
                config["reconstruction"]["rotation_axis_position"] = str(cor_nabu_ref)

        _logger.info(f"set nabu reconstruction parameters to {scan}")
        scan.nabu_recons_params = config
        res = {}
        for db in delta_beta_s:
            l_config = deepcopy(config)
            if "output" not in config:
                l_config["output"] = {}
            if output_dir is None:
                l_config["output"]["location"] = os.path.join(
                    scan.path, DEFAULT_RECONS_FOLDER
                )
            else:
                l_config["output"]["location"] = output_dir
            # TODO: allow file format modifications
            l_config["output"]["file_format"] = "hdf5"
            if "phase" not in config:
                l_config["phase"] = {}
            l_config["phase"]["delta_beta"] = db
            l_config["phase"]["method"] = "Paganin"
            res[db] = l_config
        return res

    def _run_one_slice_several_db(
        self,
        scan,
        configs,
        slice_index,
        advancement,
        dry_run,
        axis,
        cluster_config: dict | None,
    ):
        future_tomo_objs = {}
        success = True
        recons_urls = {}
        std_outs = []
        std_errs = []

        if not isinstance(cluster_config, (dict, type(None))):
            raise TypeError(
                f"'cluster_config' is expected to be a dict or None. Get {type(cluster_config)} instead."
            )

        runners = run_nabu_one_slice_several_config(
            nabu_configs=configs,
            scan=scan,
            slice_index=slice_index,
            dry_run=dry_run,
            file_format="hdf5",
            advancement=advancement,
            cluster_config=cluster_config,
            process_id=self.process_id,
            instantiate_classes_only=True,
            output_file_prefix_pattern=None,
            axis=axis,
        )

        for runner in runners:
            if self._cancelled:
                break

            self._current_processing = runner
            try:
                results = runner.run()
            except TimeoutError as e:
                _logger.error(e)
            else:
                assert isinstance(
                    results, dict
                ), "results should be a dictionary with delta-beta as key and urls as value"

                for db, res in results.items():
                    success = success and res.success
                    if isinstance(res, ResultsWithStd):
                        std_outs.append(res.std_out)
                        std_errs.append(res.std_err)
                    if (
                        isinstance(res, ResultsLocalRun)
                        and len(res.results_identifiers) > 0
                    ):
                        assert (
                            len(res.results_identifiers) == 1
                        ), "only one slice expected"
                        recons_urls[db] = res.results_identifiers[0]
                    if isinstance(res, ResultSlurmRun):
                        future_tomo_obj = FutureTomwerObject(
                            tomo_obj=scan,
                            process_requester_id=self.process_id,
                            futures=res.future_slurm_jobs,
                        )
                        future_tomo_objs[db] = future_tomo_obj

            if advancement is not None:
                advancement.update()

        return success, recons_urls, future_tomo_objs, std_outs, std_errs

    def _post_processing(
        self,
        scan,
        db_reconstructions,
    ):
        post_processing = _PostProcessing(
            scan=scan,
            db_reconstructions=db_reconstructions,
            pool_size=self.get_input_value("pool_size", self.DEFAULT_POOL_SIZE),
        )
        post_processing._cancelled = self._cancelled
        self._current_processing = post_processing
        return post_processing.run()

    def _resolve_futures(
        self,
        scan,
        nabu_config: dict,
        db_reconstructions,
        slice_index,
        axis,
        future_tomo_objs: dict,
    ):
        assert isinstance(nabu_config, dict)
        pag = False
        ctf = False
        if "phase" in nabu_config:
            phase_method = nabu_config["phase"].get("method", "").lower()
            if phase_method in ("pag", "paganin"):
                pag = True
            elif phase_method in ("ctf",):
                ctf = True

        # treat future.
        for db, future_tomo_obj in future_tomo_objs.items():
            if self._cancelled:
                break

            future_tomo_obj.results()
            if future_tomo_obj.cancelled() or future_tomo_obj.exceptions():
                continue
            file_prefix = SingleSliceRunner.get_file_basename_reconstruction(
                scan=scan,
                slice_index=slice_index,
                pag=pag,
                db=int(db) if db is not None else None,
                ctf=ctf,
                axis=axis,
            )
            # retrieve url
            volume_identifier = nabu_utils.get_recons_volume_identifier(
                file_prefix=file_prefix,
                location=nabu_config["output"]["location"],
                file_format=nabu_config.get("file_format", "hdf5"),
                scan=scan,
                slice_index=None,
                axis=axis,
            )

            assert len(volume_identifier) <= 1, "only one slice expected"
            if len(volume_identifier) == 1:
                db_reconstructions[db] = volume_identifier[0]
            else:
                _logger.warning(
                    f"something went wrong with reconstruction of {db} from {str(scan)}"
                )

    @staticmethod
    def _preprocess_slice_index(slice_index):
        if isinstance(slice_index, str):
            if not slice_index == "middle":
                raise ValueError(f"slice index {slice_index} not recognized")
            else:
                return slice_index
        elif not len(slice_index) == 1:
            raise ValueError("only manage one slice")
        else:
            return list(slice_index.values())[0]

    def _process_end(self, scan, db_res, score_rois):
        assert isinstance(scan, TomwerScanBase)
        state = ProcessManager().get_dataset_state(
            dataset_id=scan.get_identifier(), process=self
        )
        if state not in (
            DatasetState.CANCELLED,
            DatasetState.FAILED,
            DatasetState.SKIPPED,
        ):
            try:
                extra = {
                    logconfig.DOC_TITLE: self._scheme_title,
                    logconfig.SCAN_ID: str(scan),
                }
                slice_index = self.inputs.sa_delta_beta_params.get("slice_index", None)

                if db_res is None:
                    info = f"fail to compute delta/beta scores of slice {slice_index} for scan {scan}."
                    _logger.processFailed(info, extra=extra)
                    ProcessManager().notify_dataset_state(
                        dataset=scan,
                        process=self,
                        state=DatasetState.FAILED,
                        details=info,
                    )
                else:
                    info = f"delta/beta scores of slice {slice_index} for scan {scan} computed."
                    _logger.processSucceed(info, extra=extra)
                    ProcessManager().notify_dataset_state(
                        dataset=scan,
                        process=self,
                        state=DatasetState.WAIT_USER_VALIDATION,
                        details=info,
                    )
            except Exception as e:
                _logger.error(e)
            else:
                if self._dump_process:
                    self.save_results_to_disk(scan=scan)
                    if self.dump_roi:
                        self.dump_rois(scan, score_rois=score_rois)

    @staticmethod
    def dump_rois(scan, score_rois: dict):
        if not isinstance(score_rois, dict):
            raise TypeError("score_rois is expected to be a dict")

        if score_rois is None or len(score_rois) == 0:
            return

        if scan.saaxis_params.scores in (None, {}):
            return

        sa_delta_beta_results_url = SADeltaBetaTask.get_results_url(scan=scan)

        # save it to the file
        with HDF5File(sa_delta_beta_results_url.file_path(), mode="a") as h5f:
            nx_process = h5f.require_group(sa_delta_beta_results_url.data_path())
            score_roi_grp = nx_process.require_group("score_roi")
            for db, roi in score_rois.items():
                score_roi_grp[str(db)] = roi
                score_roi_grp[str(db)].attrs["interpretation"] = "image"

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "semi-automatic delta/beta finder"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Semi automatic center of rotation / axis calculation"

    @staticmethod
    def get_results_url(scan):
        return DataUrl(
            file_path=scan.get_relative_file(
                os.path.join(DEFAULT_RECONS_FOLDER, "tomwer_sadelta_beta.h5")
            ),
            data_path=scan.entry or "entry",
            scheme="silx",
        )

    @staticmethod
    def save_results_to_disk(scan):
        if scan.saaxis_params.scores in (None, {}):
            return

        results_url = SADeltaBetaTask.get_results_url(scan=scan)
        # save it to the file
        with HDF5File(results_url.file_path(), mode="w") as h5f:
            nx_process = h5f.require_group(results_url.data_path())
            if "NX_class" not in nx_process.attrs:
                nx_process.attrs["NX_class"] = "NXprocess"

            results = nx_process.require_group("results")

            for cor, (url, score) in scan.sa_delta_beta_params.scores.items():
                results_db = results.require_group(str(cor))
                for method in ScoreMethod:
                    if method is ScoreMethod.TOMO_CONSISTENCY:
                        continue
                    results_db[method.value] = score.get(method)

                link_path = os.path.relpath(
                    url.file_path(),
                    os.path.dirname(results_url.file_path()),
                )
                results_db["reconstructed_slice"] = h5py.ExternalLink(
                    link_path, url.data_path()
                )

    @staticmethod
    def load_results_from_disk(scan):
        results_url = SADeltaBetaTask.get_results_url(scan=scan)
        process_file = results_url.file_path()

        if process_file is None or not os.path.exists(process_file):
            _logger.warning(
                "Unable to find process file. Unable to read " "existing processing"
            )
            return None, None

        try:
            with EntryReader(results_url) as h5f_entry_node:
                scores = core_utils.get_scores(h5f_entry_node)
                if (
                    "results" in h5f_entry_node
                    and "delta_beta" in h5f_entry_node["results"]
                ):
                    selected = h5py_read_dataset(
                        h5f_entry_node["results"]["delta_beta"]
                    )
                else:
                    _logger.warning(f"no results found for {scan}")
                    selected = None
            return scores, selected
        except ValueError:
            _logger.warning(f"Data path ({results_url.data_path()}) not found")
            return None, None

    def cancel(self):
        """
        stop current processing
        """
        if self._current_processing is not None:
            self._cancelled = True
            self._current_processing.cancel()


class _PostProcessing:
    """class used to run SA-delta-beta post-processing on reconstructed slices"""

    DEFAULT_POOL_SIZE = 10

    def __init__(self, db_reconstructions, scan, pool_size) -> None:
        self._db_reconstructions = db_reconstructions
        self._scan = scan
        self._cancelled = False
        self.pool_size = pool_size

    @staticmethod
    def compute_score(item: tuple):
        db, (url, data), mask_disk_radius, cancelled = item
        if data is None:
            score = None
            data_roi = None
        else:
            assert data.ndim == 2
            data_roi = apply_roi(data=data, radius=mask_disk_radius, url=url)

            # move data_roi to [0-1] range
            #  preprocessing: get percentile 0 and 99 from image and
            #  "clean" highest and lowest pixels from it
            min_p, max_p = numpy.percentile(data_roi, (1, 99))
            data_roi_int = data_roi[...]
            data_roi_int[data_roi_int < min_p] = min_p
            data_roi_int[data_roi_int > max_p] = max_p
            data_roi_int = (data_roi_int - min_p) / (max_p - min_p)

            score = ComputedScore(
                tv=compute_score(data=data_roi_int, method=ScoreMethod.TV),
                std=compute_score(data=data_roi_int, method=ScoreMethod.STD),
            )
        return {db: (url, score)}, {db: data_roi}

    def run(self):
        datasets = self.load_datasets()
        assert isinstance(datasets, dict)
        mask_disk_radius = get_disk_mask_radius(datasets)
        with Pool(self.pool_size) as pool:
            res = pool.map(
                self.compute_score,
                [
                    (
                        *item,
                        mask_disk_radius,
                        self._cancelled,
                    )
                    for item in datasets.items()
                ],
            )
        scores = {}
        rois = {}
        for mydict in res:
            myscores, myrois = mydict
            scores.update(myscores)
            rois.update(myrois)
        return scores, rois

    @staticmethod
    def _load_dataset(item: tuple):
        db, volume_identifier = item
        slice_url = None
        # in case the try processing fails
        try:
            volume = VolumeFactory.create_tomo_object_from_identifier(volume_identifier)
            volumes_urls = tuple(volume.browse_data_urls())
            if len(volumes_urls) > 1:
                _logger.warning(
                    f"found a volume with mode that one url ({volumes_urls})"
                )
            slice_url = volumes_urls[0]
            data = get_data(slice_url)
        except Exception as e:
            _logger.error(
                f"Fail to compute a score for {volume_identifier}. Reason is {e}"
            )
            return {db: (slice_url, None)}
        else:
            if data.ndim == 3:
                if data.shape[0] == 1:
                    data = data.reshape(data.shape[1], data.shape[2])
                elif data.shape[2] == 1:
                    data = data.reshape(data.shape[0], data.shape[1])
                else:
                    raise ValueError(f"Data is expected to be 2D. Not {data.ndim}D")
            elif data.ndim == 2:
                pass
            else:
                raise ValueError(f"Data is expected to be 2D. Not {data.ndim}D")

            return {db: (slice_url, data)}

    def load_datasets(self):
        with Pool(self.pool_size) as pool:
            res = pool.map(
                self._load_dataset,
                self._db_reconstructions.items(),
            )
        datasets_ = {}
        for mydict in res:
            datasets_.update(mydict)

        return datasets_

    def cancel(self):
        self._cancelled = True


class SADeltaBetaProcess(SADeltaBetaTask):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        deprecated_warning(
            name="tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta.SADeltaBetaProcess",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="SADeltaBetaTask",
        )
        super().__init__(process_id, varinfo, inputs, node_id, node_attrs, execinfo)



from __future__ import annotations

import logging

import numpy
import scipy.signal
from silx.gui import qt

from tomwer.core.process.reconstruction.axis import mode as axis_mode
from tomwer.core.process.reconstruction.axis.anglemode import CorAngleMode
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils import image
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.synctools.axis import QAxisRP

from .CompareImages import CompareImages
from .AxisSettingsWidget import AxisSettingsWidget

_logger = logging.getLogger(__name__)


class AxisWidget(qt.QMainWindow):
    """
    Main widget for the computing the rotation axis.

    It contains:
    * CompareImages widget as the central widget to display two opposite radios once shifted (and flip) with the cor found
    * A control widget on the left to select the algorithm to be applied, algorithm options...

    :raises ValueError: given axis is not an instance of _QAxisRP
    """

    sigAxisEditionLocked = qt.Signal(bool)
    """Signal emitted when the status of the reconstruction parameters edition
    change"""

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the lock on the mode change"""

    sigPositionChanged = qt.Signal(tuple)
    """signal emitted when the center of rotation center change"""

    def __init__(self, axis_params, parent=None, backend=None):
        super().__init__(parent)
        if isinstance(axis_params, QAxisRP):
            self.__recons_params = axis_params
        else:
            raise TypeError("axis should be an instance of _QAxisRP")

        self._imgA = None
        self._imgB = None
        self._shiftedImgA = None
        self._flipB = True
        """Option if we want to flip the image B"""
        self._scan = None
        self._axis_params = None
        self._lastManualFlip = None
        """Cache for the last user entry for manual flip"""
        self._lastXShift = None
        # cache to know if the x shift has changed since
        self._lastYShift = None
        # cache to know if the y shift has changed
        self._lastXOrigin = None
        # cache to know if the x origin has changed since
        self._lastYOrigin = None
        # cache to know if the y origin has changed since

        self.setWindowFlags(qt.Qt.Widget)
        self._plot = CompareImages(parent=self, backend=backend)
        self._plot.setAutoResetZoom(False)
        _mode = CompareImages.VisualizationMode.COMPOSITE_A_MINUS_B
        self._plot.setVisualizationMode(_mode)
        self._plot.setAlignmentMode(CompareImages.AlignmentMode.STRETCH)
        self.setCentralWidget(self._plot)

        self._dockWidgetCtrl = qt.QDockWidget(parent=self)
        self._dockWidgetCtrl.layout().setContentsMargins(0, 0, 0, 0)
        self._dockWidgetCtrl.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._settingsWidget = AxisSettingsWidget(
            parent=self, reconsParams=self.__recons_params
        )
        self._settingsWidgetScrollArea = qt.QScrollArea(self)
        self._settingsWidgetScrollArea.setWidgetResizable(True)
        self._settingsWidgetScrollArea.setWidget(self._settingsWidget)
        self._dockWidgetCtrl.setWidget(self._settingsWidgetScrollArea)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._dockWidgetCtrl)

        # signal / slot connection
        self._settingsWidget.sigShiftChanged.connect(self._updateShift)
        self._settingsWidget.sigShiftChanged.connect(self._corChanged)
        self._settingsWidget.sigRoiChanged.connect(self._updateShift)
        self._settingsWidget.sigAuto.connect(self._updateAuto)
        self._settingsWidget.sigModeChanged.connect(self.setMode)
        self._settingsWidget.sigModeLockChanged.connect(self._modeLockChanged)
        self._settingsWidget.sigResetZoomRequested.connect(self._resetZoomPlot)
        self._settingsWidget.sigSubSamplingChanged.connect(self._updateSubSampling)
        self._settingsWidget.sigUrlChanged.connect(self._urlChanged)
        self._plot.sigCropImagesChanged.connect(self._updateShift)

        # adapt gui to the axis value
        self.setReconsParams(axis=self.__recons_params)
        self.getPlot().getPlot().setAxesDisplayed(True)

    def manual_uses_full_image(self, value):
        self._settingsWidget.manual_uses_full_image(value)

    def _modeLockChanged(self, lock):
        self.sigLockModeChanged.emit(lock)

    def _corChanged(self):
        self.sigPositionChanged.emit((self.getXShift(), self.getYShift()))

    def getPlot(self):
        return self._plot

    def _resetZoomPlot(self):
        self._plot.getPlot().resetZoom()

    def setMode(self, mode):
        """
        Define the mode to use for radio axis

        :param mode:
        :return:
        """
        mode = axis_mode.AxisMode.from_value(mode)
        with block_signals(self._settingsWidget):
            with block_signals(self._axis_params):
                self._settingsWidget.setMode(mode)
                if mode is axis_mode.AxisMode.manual:
                    self._setModeLockFrmSettings(False)

    def setEstimatedCor(self, value):
        self._settingsWidget.setEstimatedCor(value=value)

    def getEstimatedCor(self):
        return self._settingsWidget.getEstimatedCor()

    def updateXRotationAxisPixelPositionOnNewScan(self) -> bool:
        return self._settingsWidget.updateXRotationAxisPixelPositionOnNewScan()

    def _setModeLockFrmSettings(self, lock: bool):
        # only lock the push button
        with block_signals(self):
            self._settingsWidget._mainWidget._calculationWidget._lockMethodPB.setLock(
                lock
            )

    def getROIDims(self):
        if self.getMode() == axis_mode.AxisMode.manual:
            return self._settingsWidget.getROIDims()
        else:
            return None

    def getROIOrigin(self):
        if self.getMode() == axis_mode.AxisMode.manual:
            return self._settingsWidget.getROIOrigin()
        else:
            return None

    def getImgSubSampling(self):
        return self._settingsWidget.getImgSubSampling()

    def _computationRequested(self):
        self.sigComputationRequested.emit()

    def setLocked(self, locked):
        with block_signals(self):
            if self._axis_params.mode not in (axis_mode.AxisMode.manual,):
                self._axis_params.mode = axis_mode.AxisMode.manual
            self._settingsWidget.setLocked(locked)

        self.sigAxisEditionLocked.emit(locked)

    def isModeLock(self):
        return self._settingsWidget.isModeLock()

    def _validated(self):
        """callback when the validate button is activated"""
        self.sigApply.emit()

    def _setRadio2Flip(self, checked):
        self._plot.setRadio2Flip(checked)

    def _flipChanged(self, checked):
        if self.getMode() == axis_mode.AxisMode.manual:
            self._lastManualFlip = self._plot.isRadio2Flip()

        if checked == self._flipB:
            return
        else:
            self._flipB = checked
            self._updatePlot()

    def setReconsParams(self, axis: QAxisRP):
        """

        :param axis: axis to edit
        :return:
        """
        assert isinstance(axis, QAxisRP)
        self._axis_params = axis
        with block_signals(self):
            self.resetShift()
            self._settingsWidget.setAxisParams(axis)

    def setScan(self, scan):
        """
        Update the interface concerning the given scan. Try to display the
        radios for angle 0 and 180.

        :param scan: scan for which we want the axis updated.
        """
        self.clear()
        _scan = scan
        if type(scan) is str:
            try:
                _scan = ScanFactory.create_scan_object(scan)
            except ValueError:
                raise ValueError("Fail to discover a valid scan in %s" % scan)
        elif not isinstance(_scan, TomwerScanBase):
            raise ValueError(
                f"type of {scan} ({type(scan)}) is invalid, scan should be a file/dir path or an instance of ScanBase"
            )
        assert isinstance(_scan, TomwerScanBase)

        if _scan.axis_params is None:
            _scan.axis_params = QAxisRP()

        if self._scan is not None:
            self._scan.axis_params.sigAxisUrlChanged.disconnect(self._updatePlot)
        update_x_rotation_axis_pixel_position = (
            self._settingsWidget._mainWidget.updateXRotationAxisPixelPositionOnNewScan()
        )
        if (
            update_x_rotation_axis_pixel_position
            and scan.x_rotation_axis_pixel_position is not None
        ):
            self.setEstimatedCor(scan.x_rotation_axis_pixel_position)

        # update visualization
        self._scan = _scan
        self._scan.axis_params.sigAxisUrlChanged.connect(self._updatePlot)
        self._settingsWidget.setScan(scan=self._scan)
        self._updatePlot()
        self.getPlot().getPlot().resetZoom()

    def _updatePlot(self):
        if self._scan is None:
            return
        self._urlChanged()

    def _urlChanged(self):
        with block_signals(self):
            coreAngleMode = CorAngleMode.from_value(self.__recons_params.angle_mode)
            if self._scan is None:
                return
            axis_rp = self._scan.axis_params
            if coreAngleMode is CorAngleMode.manual_selection:
                manual_sel_widget = (
                    self._settingsWidget._mainWidget._inputWidget._angleModeWidget._manualFrameSelection
                )
                urls = manual_sel_widget.getFramesUrl(as_txt=False)
                axis_rp.axis_url_1, axis_rp.axis_url_2 = urls
                axis_rp.flip_lr = manual_sel_widget.isFrame2LRFLip()
            else:
                axis_rp.flip_lr = True
                res = self._scan.get_opposite_projections(mode=coreAngleMode)
                axis_rp.axis_url_1 = res[0]
                axis_rp.axis_url_2 = res[1]

            if axis_rp.n_url() < 2:
                _logger.error("Fail to detect radio for axis calculation")
            elif axis_rp.axis_url_1.url:
                # if necessary normalize data
                axis_rp.axis_url_1.normalize_data(self._scan, log_=False)
                axis_rp.axis_url_2.normalize_data(self._scan, log_=False)

                paganin = self.__recons_params.paganin_preproc
                # check if normed
                if paganin:
                    imgA = axis_rp.axis_url_1.normalized_data_paganin
                    imgB = axis_rp.axis_url_2.normalized_data_paganin
                else:
                    imgA = axis_rp.axis_url_1.normalized_data
                    imgB = axis_rp.axis_url_2.normalized_data
                assert imgA is not None
                assert imgB is not None
                self.setImages(imgA=imgA, imgB=imgB, flipB=axis_rp.flip_lr)
            else:
                _logger.error(
                    "fail to find radios for angle 0 and 180. Unable to update axis gui"
                )

    def clear(self):
        if self._scan is not None:
            self._scan.axis_params.sigAxisUrlChanged.disconnect(self._updatePlot)
        self._scan = None

    def setImages(self, imgA: numpy.array, imgB: numpy.array, flipB: bool):
        """

        :warning: does not reset the shift when change images

        :param imgA: first image to compare. Will be the one shifted
        :param imgB: second image to compare
        :param flipB: True if the image B has to be flipped
        """
        assert imgA is not None
        assert imgB is not None
        _imgA = imgA
        _imgB = imgB

        if _imgA.shape != _imgB.shape:
            _logger.error(
                "The two provided images have incoherent shapes "
                f"({_imgA.shape} vs {_imgB.shape})"
            )
        elif _imgA.ndim != 2:
            _logger.error("Image shape are not 2 dimensional")
        else:
            self._imgA = _imgA
            self._imgB = _imgB
            self._flipB = flipB

            self._settingsWidget._roiControl.setLimits(
                width=self._imgA.shape[1], height=self._imgA.shape[0]
            )
            self._updateShift()

    def _updateSubSampling(self):
        self._updateShift()
        self.getPlot().getPlot().resetZoom()

    def _updateShift(self, xShift=None, yShift=None):
        if self._imgA is None or self._imgB is None:
            return
        xShift = xShift or self.getXShift()
        yShift = yShift or self.getYShift()

        # TODO: we might avoid flipping image at each new x_shift...
        _imgA, _imgB = self._getRawImages()
        # apply shift
        if xShift == 0.0 and yShift == 0.0:
            self._shiftedImgA = _imgA
            self._shiftedImgB = _imgB
        else:
            try:
                cval_imgA = _imgA.min()
                cval_imgB = _imgB.min()
            except ValueError:
                _logger.warning("enable to retrieve imgA.min() and / or" "imgB.min().")
                cval_imgA = 0
                cval_imgB = 0
            try:
                x_shift = self.getXShift() / self.getImgSubSampling()
                y_shift = self.getYShift() / self.getImgSubSampling()
                self._shiftedImgA = image.shift_img(
                    data=_imgA,
                    dx=-x_shift,
                    dy=y_shift,
                    cval=cval_imgA,
                )
                self._shiftedImgB = image.shift_img(
                    data=_imgB,
                    dx=x_shift,
                    dy=y_shift,
                    cval=cval_imgB,
                )
                crop = self.getPlot().cropComparedImages()

                if not crop:
                    # handling of the crop:
                    # 1. we will concatenate the shifted array with the unshifted to avoid crop
                    # 2. in order to handled properly the shift and overlaps we need to add an empty array
                    abs_x_shift = abs(int(x_shift))
                    buffer_array_img_A = numpy.full(
                        shape=(self._shiftedImgA.shape[0], abs_x_shift),
                        fill_value=cval_imgA,
                    )
                    buffer_array_img_B = numpy.full(
                        shape=(self._shiftedImgB.shape[0], abs_x_shift),
                        fill_value=cval_imgB,
                    )
                    if x_shift == 0:
                        pass
                    elif x_shift > 0:
                        self._shiftedImgA = numpy.concatenate(
                            (
                                _imgA[:, :abs_x_shift],
                                self._shiftedImgA,
                                buffer_array_img_A,
                            ),
                            axis=1,
                        )
                        self._shiftedImgB = numpy.concatenate(
                            (
                                buffer_array_img_B,
                                self._shiftedImgB,
                                _imgB[:, -abs_x_shift:],
                            ),
                            axis=1,
                        )
                    else:
                        self._shiftedImgA = numpy.concatenate(
                            (
                                buffer_array_img_A,
                                self._shiftedImgA,
                                _imgA[:, :abs_x_shift],
                            ),
                            axis=1,
                        )
                        self._shiftedImgB = numpy.concatenate(
                            (
                                _imgB[:, :abs_x_shift],
                                self._shiftedImgB,
                                buffer_array_img_B,
                            ),
                            axis=1,
                        )
            except ValueError as e:
                _logger.error(e)
                self._shiftedImgA = _imgA
                self._shiftedImgB = _imgB

        with block_signals(self):
            try:
                self._plot.setData(
                    image1=self._shiftedImgA,
                    image2=self._shiftedImgB,
                )
            except ValueError:
                _logger.warning(
                    "Unable to set images. Maybe there is some "
                    "incomplete dataset or an issue with "
                    "normalization."
                )
            roi_origin = self.getROIOrigin()
            if roi_origin is not None:
                x_origin, y_origin = roi_origin
            else:
                x_origin = y_origin = None
            self._lastXShift = xShift
            self._lastYShift = yShift
            self._lastXOrigin = x_origin
            self._lastYOrigin = y_origin

    def _getRawImages(self):
        def selectROI(data, width, height, x_origin, y_origin, subsampling):
            assert subsampling > 0
            x_min = x_origin - width // 2
            x_max = x_origin + width // 2
            y_min = y_origin - height // 2
            y_max = y_origin + height // 2
            return data[y_min:y_max:subsampling, x_min:x_max:subsampling]

        # get images and apply ROI if any
        _roi_dims = self.getROIDims()
        _origin = self.getROIOrigin()
        subsampling = self.getImgSubSampling()
        _imgA = self._imgA
        _imgB = self._imgB
        # flip image B
        _imgB = numpy.fliplr(_imgB) if self._flipB else _imgB
        if _roi_dims is not None:
            assert type(_roi_dims) is tuple, f"invalide roi value {_roi_dims}"
            _imgA = selectROI(
                _imgA,
                width=_roi_dims[0],
                height=_roi_dims[1],
                x_origin=_origin[0],
                y_origin=_origin[1],
                subsampling=subsampling,
            )
            _imgB = selectROI(
                _imgB,
                width=_roi_dims[0],
                height=_roi_dims[1],
                x_origin=_origin[0],
                y_origin=_origin[1],
                subsampling=subsampling,
            )
        return _imgA, _imgB

    def _updateAuto(self):
        _imgA, _imgB = self._getRawImages()
        correlation = scipy.signal.correlate2d(in1=_imgA, in2=_imgB)
        y, x = numpy.unravel_index(numpy.argmax(correlation), correlation.shape)
        self._setShift(x=x, y=y)

    def resetShift(self):
        with block_signals(self._settingsWidget):
            self._settingsWidget.reset()
        if self._imgA is not None and self._imgB is not None:
            self.setImages(imgA=self._imgA, imgB=self._imgB, flipB=self._flipB)

    # expose API

    def getXShift(self):
        return self._settingsWidget.getXShift()

    def setXShift(self, x):
        self._settingsWidget.setXShift(x=x)

    def getYShift(self):
        return self._settingsWidget.getYShift()

    def setYShift(self, y):
        self._settingsWidget.setYShift(y=y)

    def _setShift(self, x, y):
        self._settingsWidget.setShift(x, y)

    def getShiftStep(self):
        return self._settingsWidget.getShiftStep()

    def setShiftStep(self, value):
        self._settingsWidget.setShiftStep(value=value)

    def getAxisParams(self):
        return self._settingsWidget.getAxisParams()

    def getMode(self):
        return self._settingsWidget.getMode()

    def setModeLock(self, mode):
        return self._settingsWidget.setModeLock(mode=mode)

    def isYAxisInverted(self) -> bool:
        return self._settingsWidget.isYAxisInverted()

    def setYAxisInverted(self, checked: bool):
        return self._settingsWidget.setYAxisInverted(checked=checked)



import gc
import pickle
import shutil
import tempfile
import time
import os

import pytest
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.reconstruction.AxisOW import AxisOW
from tomwer.core.utils.lbsram import mock_low_memory
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.settings import mock_lsbram
from tomwer.core.utils.scanutils import MockNXtomo, MockEDF
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestOWAxis(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        self._window = AxisOW()
        self.recons_params = self._window.recons_params
        self.tempdir = tempfile.mkdtemp()
        self.scan = MockNXtomo(
            scan_path=os.path.join(self.tempdir, "nx_tomo.nx"),
            n_proj=10,
            n_ini_proj=10,
            scan_range=180,
            dim=20,
            energy=12.3,
        ).scan

        self._window.getAxisParams().mode = AxisMode.centered
        self._window.show()
        self.qWaitForWindowExposed(self._window)

    def tearDown(self):
        shutil.rmtree(self.tempdir)
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        gc.collect()

    def testAxisLock(self):
        """Test behavior when locking the axis position. Could not be included
        in the tomwer/gui because the lock action is only available for the OW
        """
        assert self._window.getAxisParams().mode in (
            AxisMode.centered,
            AxisMode.global_,
        )
        radio_axis = self._window._widget._axisWidget
        main_widget = self._window._widget
        self.assertFalse(main_widget._controlWidget._lockBut.isLocked())
        self.assertTrue(radio_axis._settingsWidget._mainWidget.isEnabled())
        self.mouseClick(main_widget._controlWidget._lockBut, qt.Qt.LeftButton)
        self.qapp.processEvents()
        self.assertTrue(main_widget._controlWidget._lockBut.isLocked())
        # when the lock button is activated we should automatically switch to
        # the manual mode
        self.assertTrue(self._window.getAxisParams().mode is AxisMode.manual)
        self.assertFalse(radio_axis._settingsWidget._mainWidget.isEnabled())

    def test_serializing(self):
        pickle.dumps(self._window.recons_params.to_dict())

    def test_literal_dumps(self):
        literal_dumps(self._window.recons_params.to_dict())


class TestWindowAxisComputation(TestCaseQt):
    @staticmethod
    def _long_computation(scan):
        time.sleep(5)
        return -1

    """Test that the axis widget work correctly"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self.tempdir = tempfile.mkdtemp()
        self._mainWindow = AxisOW()
        self.recons_params = self._mainWindow.recons_params
        self._window = self._mainWindow._widget
        self.scan = MockNXtomo(
            scan_path=os.path.join(self.tempdir, "nx_tomo.nx"),
            n_proj=10,
            n_ini_proj=10,
            scan_range=180,
            dim=20,
            energy=12.3,
        ).scan
        self._mainWindow.show()
        self.qWaitForWindowExposed(self._mainWindow)
        self._mainWindow.setMode("manual")

    def tearDown(self):
        shutil.rmtree(self.tempdir)
        self._mainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._mainWindow.close()
        self._mainWindow = None
        self._window = None
        gc.collect()

    def testComputationSucceed(self):
        """Test gui if the axis position is correctly computed"""
        self.recons_params.mode = AxisMode.manual
        radio_axis = self._window._axisWidget
        self.assertEqual(radio_axis.getMode(), AxisMode.manual)
        radio_axis.setXShift(2.345)
        self._mainWindow.process(self.scan)
        self.qapp.processEvents()
        self.assertEqual(radio_axis.getMode(), AxisMode.manual)
        self.assertEqual(self.recons_params.relative_cor_value, 2.345)
        position_info_widget = self._window._controlWidget._positionInfo
        self.assertEqual(position_info_widget._relativePositionQLE.text(), "2.345")


global _computation_res
_computation_res = 0


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestAxisStack(TestCaseQt):
    """Test axis computation of a stack of scan"""

    @staticmethod
    def _test_computation(scan):
        global _computation_res
        _computation_res += 1
        return _computation_res

    def setUp(self):
        # not working due to OW
        TestCaseQt.setUp(self)
        self._scan1 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._scan2 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._scan3 = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        self._mainWindow = AxisOW()
        self.recons_params = self._mainWindow._axis_params
        self._mainWindow._skip_exec(True)

        global _computation_res
        _computation_res = 0

        self._mainWindow.patch_calc_method(
            AxisMode.centered, TestAxisStack._test_computation
        )
        self.recons_params.mode = AxisMode.centered

        self._mainWindow.show()
        self.qWaitForWindowExposed(self._mainWindow)

    def tearDown(self):
        mock_low_memory(False)
        mock_lsbram(False)
        self._mainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._mainWindow.close()
        self._mainWindow = None
        self._recons_params = None
        self.qapp.processEvents()
        for scan in (self._scan1, self._scan2, self._scan3):
            shutil.rmtree(scan.path)
        gc.collect()

    def testLowMemory(self):
        """Make sure the axis computation will be skip if we are in low memory"""
        mock_low_memory(True)
        mock_lsbram(True)
        for scan in (self._scan1, self._scan2, self._scan3):
            self.qapp.processEvents()
            self._mainWindow.new_data_in(scan)

        self.assertEqual(self._scan1.axis_params, None)
        self.assertEqual(self._scan2.axis_params, None)
        self.assertEqual(self._scan3.axis_params, None)

    def testLockStack(self):
        """Check that axis position will be simply copy if we are in a lock
        stack"""
        self.recons_params.mode = AxisMode.manual
        position_value = 0.36
        self.recons_params.set_relative_value(position_value)

        for scan in (self._scan1, self._scan2, self._scan3):
            self._mainWindow.process(scan)

        for i in range(5):
            self.qapp.processEvents()
            time.sleep(0.2)
            self.qapp.processEvents()

        self.assertEqual(self._scan1.axis_params.relative_cor_value, position_value)
        self.assertEqual(self._scan2.axis_params.relative_cor_value, position_value)
        self.assertEqual(self._scan3.axis_params.relative_cor_value, position_value)



# coding: utf-8
from __future__ import annotations

import gc
import logging
import os
import shutil
import tempfile
import time

import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.reconstruction.DarkRefAndCopyOW import (
    DarkRefAndCopyOW,
)
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefsTask
from tomwer.core.process.reconstruction.darkref.params import ReduceMethod as DkrfMethod
from tomwer.core.process.reconstruction.darkref.settings import (
    DARKHST_PREFIX,
    REFHST_PREFIX,
)
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.utils.scanutils import MockEDF, MockNXtomo
from tomwer.synctools.darkref import QDKRFRP

logging.disable(logging.INFO)


class TestEDFDarkRefWidget(TestCaseQt):
    """class testing the DarkRefWidget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._recons_params = QDKRFRP()
        self.widget = DarkRefAndCopyOW(parent=None, reconsparams=self._recons_params)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        self.qapp.processEvents()
        gc.collect()

    def testSyncRead(self):
        """Make sure any modification on the self._reconsParams is
        applied on the GUI"""
        rp = self._recons_params
        self.assertTrue(rp["REFSRMV"] is False)
        self.assertFalse(
            self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        )
        rp["REFSRMV"] = True
        self.assertTrue(
            self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        )

        pattern = self.widget.widget.mainWidget.tabExpert._refLE.text()
        newText = "popo.*"
        assert pattern != newText
        rp["RFFILE"] = newText
        self.assertTrue(
            self.widget.widget.mainWidget.tabExpert._refLE.text() == newText
        )

    def testSyncWrite(self):
        """Test that if we edit through the :class:`DarkRefWidget` then the
        modification are fall back into the self._reconsParams"""
        rp = self._recons_params

        # test patterns
        pattern = self.widget.widget.mainWidget.tabExpert._refLE.text()
        newText = "popo.*"
        assert pattern != newText
        self.widget.widget.mainWidget.tabExpert._refLE.setText(newText)
        self.widget.widget.mainWidget.tabExpert._refLE.editingFinished.emit()
        qt.QApplication.instance().processEvents()
        self.assertTrue(rp["RFFILE"] == newText)
        self.widget.widget.mainWidget.tabExpert._darkLE.setText(newText)
        self.widget.widget.mainWidget.tabExpert._darkLE.editingFinished.emit()
        qt.QApplication.instance().processEvents()
        self.assertTrue(rp["DKFILE"] == newText)

        # test calc mode
        self.widget.widget.mainWidget.tabGeneral._darkWCB.setMode(DkrfMethod.NONE)
        self.widget.widget.mainWidget.tabGeneral._refWCB.setMode(DkrfMethod.MEDIAN)
        self.assertTrue(rp["DARKCAL"] == DkrfMethod.NONE)
        self.assertTrue(rp["REFSCAL"] == DkrfMethod.MEDIAN)

        # test options
        cuRm = self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        self.widget.widget.mainWidget.tabGeneral._rmOptionCB.setChecked(not cuRm)
        self.assertTrue(rp["REFSRMV"] == (not cuRm))
        self.assertTrue(rp["DARKRMV"] == (not cuRm))

        cuSkip = self.widget.widget.mainWidget.tabGeneral._skipOptionCB.isChecked()
        self.widget.widget.mainWidget.tabGeneral._skipOptionCB.setChecked(not cuSkip)
        # warning : here value of skip and overwrite are of course inverse
        self.assertTrue(rp["DARKOVE"] == cuSkip)
        self.assertTrue(rp["REFSOVE"] == cuSkip)


@pytest.mark.skip("Fail on CI")
class TestDarkRefCopyWithEDFAndHDF5(TestCaseQt):
    """Test the DarkRefCopy orange widget behaviour"""

    def setUp(self) -> None:
        TestCaseQt.setUp(self)
        self._folder = tempfile.mkdtemp()

        # define scans to be treated
        hdf5_mock_with_refs = MockNXtomo(
            scan_path=os.path.join(self._folder, "h5_with_refs"),
            n_proj=10,
            n_ini_proj=10,
            dim=20,
            create_ini_flat=True,
            create_ini_dark=True,
            create_final_flat=False,
        )
        self.hdf5_acquisition_with_refs = hdf5_mock_with_refs.scan

        hdf5_mock_without_refs = MockNXtomo(
            scan_path=os.path.join(self._folder, "h5_without_refs"),
            n_proj=10,
            n_ini_proj=10,
            dim=20,
            create_ini_flat=False,
            create_ini_dark=False,
        )
        self.hdf5_acquisition_without_refs = hdf5_mock_without_refs.scan

        hdf5_mock_without_refs_incoherent_dim = MockNXtomo(
            scan_path=os.path.join(self._folder, "h5_without_refs_different_dim"),
            n_proj=10,
            n_ini_proj=10,
            dim=21,
            create_ini_flat=False,
            create_ini_dark=False,
        )
        self.hdf5_acquisition_without_refs_incoherent_dim = (
            hdf5_mock_without_refs_incoherent_dim.scan
        )

        edf_mock_without_ref = MockEDF(
            scan_path=os.path.join(self._folder, "edf_without_refs"), dim=20, n_radio=20
        )
        self.edf_acquisition_without_ref = EDFTomoScan(
            scan=edf_mock_without_ref.scan_path
        )

        # processes set up
        self._recons_params = QDKRFRP()
        self.widget = DarkRefAndCopyOW(parent=None, reconsparams=self._recons_params)
        self.widget.setForceSync(True)
        self.widget.show()

    def tearDown(self) -> None:
        time.sleep(0.5)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self._folder)
        gc.collect()

    def testCopyInactive(self):
        self.widget.setCopyActive(False)

        self.widget.process(self.hdf5_acquisition_with_refs)

        # make sure no dark or flats exists for the hdf5 without refs
        for scan in (
            self.hdf5_acquisition_without_refs_incoherent_dim,
            self.hdf5_acquisition_without_refs,
        ):
            self.widget.process(scan)

        # make sure no dark or flats exists for the edf one
        self.widget.process(self.edf_acquisition_without_ref)

    def testCopyActive(self):
        self.widget.setCopyActive(True)
        self.widget.setModeAuto(True)

        self.widget.process(self.hdf5_acquisition_with_refs)
        # 1. make sure dark has been processed for the one with ref
        self.assertTrue(self.widget.hasDarkStored())
        self.assertTrue(self.widget.hasFlatStored())

        # 2. make sure copy has been processed for the the 'compatible hdf5'
        self.widget.process(self.hdf5_acquisition_without_refs)

        # 3. make sure copy has been processed for the the 'compatible edf'
        self.widget.process(self.edf_acquisition_without_ref)
        self.assertEqual(
            len(
                DarkRefsTask.getRefHSTFiles(
                    self.edf_acquisition_without_ref.path, prefix=REFHST_PREFIX
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                DarkRefsTask.getDarkHSTFiles(
                    self.edf_acquisition_without_ref.path, prefix=DARKHST_PREFIX
                )
            ),
            1,
        )

        # 4. make sure process but no copy made if incompatible size
        self.widget.process(self.hdf5_acquisition_without_refs_incoherent_dim)



import logging
import os

import numpy
import pytest
from silx.gui import qt
from tomwer.core.volume.edfvolume import EDFVolume
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.process.reconstruction.output import (
    PROCESS_FOLDER_RECONSTRUCTED_SLICES,
)
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.stacks import RadioStack, SliceStack
from tomwer.tests.utils import skip_gui_test
from tomwer.tests.conftest import qtapp  # noqa F401

logging.disable(logging.INFO)


@pytest.mark.parametrize("use_identifiers", (True, False))
@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
def test_slice_stack(
    qtapp,  # noqa F811
    tmp_path,
    use_identifiers,
):
    """
    :param use_identifiers: If true will add the different tomo object using identifiers instead of class instances.
    """
    widget = SliceStack()
    widget.setAttribute(qt.Qt.WA_DeleteOnClose)

    url_list = widget._viewer._urlsTable._urlsTable
    assert url_list.count() == 0
    n_scan = 5

    root_dir = tmp_path / "test_slice_stack"
    root_dir.mkdir()

    def create_vol_data(n_slice=1):
        return numpy.random.random(10 * 10 * n_slice).reshape(n_slice, 10, 10)

    volume_metadata = {
        "nabu_config": {
            "reconstruction": {
                "method": "FBP",
            },
            "phase": {
                "method": "paganin",
                "delta_beta": 110.0,
            },
        },
        "processing_options": {
            "reconstruction": {
                "voxel_size_cm": (0.2, 0.2, 0.2),
                "rotation_axis_position": 104,
                "enable_halftomo": True,
                "fbp_filter_type": "Hilbert",
                "sample_detector_dist": 0.4,
            },
            "take_log": {
                "log_min_clip": 1.0,
                "log_max_clip": 10.0,
            },
        },
    }

    # step 1: test adding some scans
    for i_scan in range(n_scan):
        scan = MockNXtomo(
            scan_path=os.path.join(root_dir, f"scan{i_scan}"),
            n_proj=10,
            n_ini_proj=10,
            scan_range=180,
            dim=10,
        ).scan
        volume = EDFVolume(
            folder=os.path.join(
                scan.path,
                PROCESS_FOLDER_RECONSTRUCTED_SLICES,
                f"edf_volume_scan_{i_scan}",
            ),
            volume_basename=f"edf_volume_scan_{i_scan}",
            data=create_vol_data(),
            metadata=volume_metadata,
        )
        volume.save()
        scan.set_latest_reconstructions((volume,))
        assert len(scan.latest_reconstructions) == 1
        if use_identifiers:
            widget.addTomoObj(scan.get_identifier().to_str())
        else:
            widget.addTomoObj(scan)

    last_scan_added = scan

    assert url_list.count() == n_scan
    assert len(widget._volume_id_to_urls) == n_scan

    # step 2: test adding some volumes
    n_volume = 4
    last_edf_volume_added = None
    last_hdf5_volume_added = None

    for i_vol in range(n_volume):
        volume_metadata["nabu_config"]["phase"]["delta_beta"] += i_vol
        volume_type = [EDFVolume, HDF5Volume][i_vol % 2]
        if volume_type == EDFVolume:
            volume = EDFVolume(
                folder=root_dir
                / PROCESS_FOLDER_RECONSTRUCTED_SLICES
                / f"slice_{i_vol}",
                volume_basename=f"slice_{i_vol}",
                data=create_vol_data(),
                metadata=volume_metadata,
            )
            last_edf_volume_added = volume
        elif volume_type == HDF5Volume:
            volume = HDF5Volume(
                file_path=os.path.join(
                    root_dir
                    / PROCESS_FOLDER_RECONSTRUCTED_SLICES
                    / f"my_vol_{i_vol}.hdf5"
                ),
                data_path="entry0000",
                data=create_vol_data(),
                metadata=volume_metadata,
            )
            last_hdf5_volume_added = volume
        else:
            raise TypeError
        volume.save()
        if use_identifiers:
            widget.addTomoObj(volume.get_identifier().to_str())
        else:
            widget.addTomoObj(volume)

    assert len(widget._volume_id_to_urls) == (n_scan + n_volume)
    assert len(widget._viewer.getUrls()) == (n_scan + n_volume)

    widget.clear()
    assert len(widget._volume_id_to_urls) == 0

    # check adding objects from a list of string
    assert n_volume >= 2, "following test won't be able to be correctly executed"
    widget._addTomoObjectsFromStrList(
        [
            last_hdf5_volume_added.file_path,
            last_edf_volume_added.url.file_path(),
            last_scan_added.path,
        ]
    )
    # dummy check metadata are displayed
    widget._viewer._reconsWidget._deltaBetaQLE.text().startswith("1")
    assert len(widget._volume_id_to_urls) == 3
    # try adding a non existing path
    widget._addTomoObjectsFromStrList(
        [
            "/none/existing/path",
            "/none/existing/path.hdf5",
        ]
    )

    widget.close()


@pytest.mark.parametrize("use_identifiers", (True, False))
@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
def test_radio_stack(
    qtapp,  # noqa F811
    tmp_path,
    use_identifiers: bool,
):
    """
    :param use_identifiers: If true will add the different tomo object using identifiers instead of class instances.
    """
    widget = RadioStack()
    widget.setAttribute(qt.Qt.WA_DeleteOnClose)

    widget_urls = widget._viewer._urlsTable._urlsTable
    n_scan = 3

    root_dir = tmp_path / "test_radio_stack"
    root_dir.mkdir()
    proj_per_scan = 10

    # step 1: test adding some scans
    for i_scan in range(n_scan):
        scan = MockNXtomo(
            scan_path=os.path.join(root_dir, f"scan{i_scan}"),
            n_proj=proj_per_scan,
            n_ini_proj=proj_per_scan,
            scan_range=180,
            dim=10,
        ).scan
        if use_identifiers:
            widget.addTomoObj(scan.get_identifier().to_str())
        else:
            widget.addTomoObj(scan)
    assert widget_urls.count() == n_scan * proj_per_scan

    widget.show()
    widget.close()



import numpy

from tomwer.core.process.reconstruction.utils.cor import (
    absolute_pos_to_relative,
    relative_pos_to_absolute,
)


def test_cor_conversion():
    """
    test absolute_pos_to_relative and relative_pos_to_absolute functions
    """
    assert relative_pos_to_absolute(relative_pos=0.0, det_width=100) == 50.0
    assert relative_pos_to_absolute(relative_pos=0.0, det_width=101) == 50.5

    assert absolute_pos_to_relative(absolute_pos=20, det_width=500) == -230.0
    assert absolute_pos_to_relative(absolute_pos=300, det_width=500) == 50.0

    for det_width in (10, 20, 30, 50):
        for relative_cor_pos in (0, -2.3, -4.5):
            numpy.testing.assert_almost_equal(
                relative_cor_pos,
                absolute_pos_to_relative(
                    absolute_pos=relative_pos_to_absolute(
                        relative_pos=relative_cor_pos, det_width=det_width
                    ),
                    det_width=det_width,
                ),
            )



from __future__ import annotations

import logging
import os
import threading
import time

from ewokscore.task import Task as EwoksTask

from tomwer.core.process.control.datawatcher.datawatcherobserver import (
    _DataWatcherObserver,
    _OngoingObservation,
)
from tomwer.core.process.task import BaseProcessInfo
from tomwer.core.process.utils import LastReceivedScansDict
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.settings import get_dest_path, get_lbsram_path
from tomwer.core.signal import Signal
from tomwer.core.utils import logconfig

from .status import *  # noqa F403

logger = logging.getLogger(__name__)


class DataWatcherEwoksTask(EwoksTask, input_names=["data"], output_names=["data"]):
    """
    For now the data watcher is a 'special case'. Because it will trigger downstream workflows / prcessing each time it discover
    a new scan.
    To move to 'default' ewoks workflows. Like launching it from the command line the simpler for now is to
    simply pass the `data` input that user can fill manually.
    """

    def run(self):
        self.outputs.data = self.inputs.data


class _DataWatcher(BaseProcessInfo):
    # TODO: change: this is not properly a Task
    """DataWatcher is the class use to mange observation of scans

    It basically have one DataWatcherObserver thread which is dealing with the
        observations. It parse all folder contained under the RootDir.

    This DataWatcherObserver is run every `maxWaitBtwObsLoop`. The call to the
    DataWatcherObserver is managed by the `loopObservationThread` which is calling
    back `_launchObservation`

    DataWatcher can run infinite search or stop observation after the discovery
    of the first valid scan (`setInfiniteIteration`)
    """

    folderObserved = None

    NB_STORED_LAST_FOUND = 20
    """All found acquisition are stored until we reach this number. In this
    case the oldest one will be removed"""

    DEFAULT_OBS_METH = DET_END_XML  # noqa F405

    def __init__(self):
        super().__init__()
        """use to know if we want to continue observation when the finished
        scan is found ?"""
        self.lastFoundScans = LastReceivedScansDict(self.NB_STORED_LAST_FOUND)
        self.isObserving = False
        self._initClass()

        # get ready observation
        self._setCurrentStatus("not processing")
        self._launchObservation()
        self._checkThread = None
        self._serialize_output_data = False

    def set_serialize_output_data(self, serialize: bool):
        self._serialize_output_data = serialize

    def _launchObservation(self):
        """Main function of the widget"""
        raise NotImplementedError("Base class")

    def _initClass(self):
        self.srcPattern = get_lbsram_path()
        self.destPattern = get_dest_path()

        self.observationThread = None
        """Thread used to parse directories and found where some
        :class`DataWatcherFixObserver` have to be launcher"""
        self.obsThIsConnected = False
        self.maxWaitBtwObsLoop = 5
        """time between two observation loop, in sec"""
        self.obsMethod = self.DEFAULT_OBS_METH
        """Pattern to look in the acquisition file to know of the acquisition
         is ended or not"""
        self.startByOldest = False
        self.linux_filter_file_pattern = None

    def _switchObservation(self):
        """Switch the status of observation"""
        if self.isObserving is True:
            self.stop()
        else:
            self.start()

    def setObsMethod(self, obsMethod):
        """Set the observation method to follow.

        .. Note:: For now this will be apply on the next observation iteration.
                  We don't wan't to stop and restart an observation as sometimes
                  It can invoke a lot of start/stop if the user is editing the
                  pattern of the file for example. But this might evolve in the
                  future.
        """
        assert type(obsMethod) in (str, tuple)
        if type(obsMethod) is tuple:
            assert len(obsMethod) == 1 or (
                type(obsMethod[1]) is dict and len(obsMethod) == 2
            )
            obsMethod = obsMethod[0]
        self.obsMethod = obsMethod
        if self.isObserving:
            self._restartObservation()

    def setObservation(self, b):
        """Start a new observation (if none running ) or stop the current
        observation

        :param b: the value to set to the observation
        """
        self.start() if b else self.stop()

    def stop(self, sucess=False):
        """
        Stop the thread of observation

        :param sucess: if True this mean that we are stopping the
                            observation because we found an acquisition
                            finished. In this case we don't want to update the
                            status and the log message

        """
        if self.isObserving is False:
            return False

        self._setIsObserving(False)
        if self.observationThread is not None:
            # remove connection
            self.observationThread.quitEvent.set()
            if self.observationThread.is_alive():
                self.observationThread.join()
            if self._checkThread and self._checkThread.is_alive():
                self._checkThread.join()
            self.observationThread = None

        if sucess is False:
            self._setCurrentStatus(str("not processing"))
        logger.info("stop observation from %s" % self.folderObserved)

        return True

    def start(self):
        """Start the thread of observation

        :return bool: True if the observation was started. Otherwise this
           mean that an observation was already running
        """
        if self.isObserving is True:
            return False
        else:
            logger.info("start observation from %s" % self.folderObserved)
            self._setIsObserving(True)
            self._setCurrentStatus("not processing")
            self._launchObservation()

            return True

    def _setIsObserving(self, b):
        self.isObserving = b

    def resetStatus(self):
        """
        Reset the status to not processing. Needed to restart observation,
        like when the folder is changing
        """
        self._setCurrentStatus("not processing")

    def getFolderObserved(self):
        return self.folderObserved

    def getLinuxFilePattern(self):
        return self.linux_filter_file_pattern

    def setLinuxFilePattern(self, pattern: str | None):
        if pattern is None:
            self.linux_filter_file_pattern = pattern
        else:
            pattern = pattern.lstrip(" ").rstrip(" ")
            if pattern == "":
                self.linux_filter_file_pattern = None
            else:
                self.linux_filter_file_pattern = pattern

    def setFolderObserved(self, path):
        assert type(path) is str
        if not os.path.isdir(path):
            warning = "Can't set the observe folder to ", path, " invalid path"
            logger.warning(warning, extra={logconfig.DOC_TITLE: self._scheme_title})
        else:
            self.folderObserved = os.path.abspath(path)

    def getTimeBreakBetweenObservation(self):
        """

        :return: the duration of the break we want to do between two
            observations (in sec)
        """
        return self.maxWaitBtwObsLoop

    def setWaitTimeBtwLoop(self, time):
        if not time > 0:
            err = "invalid time given %s" % time
            raise ValueError(err)
        self.maxWaitBtwObsLoop = time

    def setStartByOldest(self, b):
        """
        When parsing folders, should we start by the oldest or the newest file

        :param b: if True, will parse folders from the oldest one
        """
        self.startByOldest = b

    def setSrcAndDestPattern(self, srcPattern: str | None, destPattern: str | None):
        """Set the values of source pattern and dest pattern

        :param srcPattern: the value to set to the source pattern
                                       (see datawatcher)
        :param destPattern: the value to set to the destination
                                        pattern (see datawatcher)
        """
        self.srcPattern = srcPattern
        self.destPattern = destPattern

    def _initObservation(self):
        """Init the thread running the data watcher functions"""
        if self.observationThread is None:
            self._createDataWatcher()
        headDir = self.getFolderObserved()
        if headDir is None or not os.path.isdir(headDir):
            self._messageNotDir(headDir)
            self.stop()
            return False

        # update information on the head folder and the start by the oldest
        self.observationThread.setHeadFolder(headDir)
        self.observationThread.setObservationMethod(self.obsMethod)
        self.observationThread.setUnixFileNamePattern(self.linux_filter_file_pattern)

        return True

    def _messageNotDir(self, dir_):
        message = "Given path (%s) isn't a directory." % dir_
        logger.warning(message, extra={logconfig.DOC_TITLE: self._scheme_title})

    def _createDataWatcher(self):
        self.observationThread = _DataWatcherObserver(
            observationClass=self._getObservationClass(),
            obsMethod=self.obsMethod,
            time_between_loops=self.maxWaitBtwObsLoop,
            srcPattern=self.srcPattern,
            destPattern=self.destPattern,
        )

    def _observation_thread_running(self):
        return self.observationThread is not None and self.observationThread.is_alive()

    def _init_check_finished_scan(self):
        self._checkThread = threading.Thread(target=self._check_finished_scan)
        self._checkThread.start()

    def _getObservationClass(self):
        return _OngoingObservation

    def _check_finished_scan(self):
        """check scanReady event"""
        while self.isObserving is True:
            if self.observationThread is not None:
                self.observationThread._check_scans_ready()
                with self.observationThread.lock:
                    if self.observationThread.scanReadyEvent.isSet():
                        for scanID in self.observationThread.latestScanReady:
                            try:
                                _scan = ScanFactory.create_scan_object(scan_path=scanID)
                            except Exception as e:
                                logger.error(
                                    f"Fail to create TomoBase from {scanID}. Error is {e}"
                                )
                            else:
                                self._signalScanReady(scan=_scan)
                        self.observationThread.scanReadyEvent.clear()
            time.sleep(0.05)

    def _restartObservation(self):
        """Reset system to launch a new observation"""
        if self.observationThread is not None:
            self.observationThread.quit()
            self._setCurrentStatus("not processing")
            self._launchObservation()

    def _statusChanged(self, status):
        assert status[0] in OBSERVATION_STATUS  # noqa 320
        self._setCurrentStatus(status[0], status[1] if len(status) == 2 else None)

    def informationReceived(self, info):
        logger.info(info)

    def _setCurrentStatus(self, status, info=None):
        """Change the current status to status"""
        assert type(status) is str
        assert status in OBSERVATION_STATUS  # noqa 320
        self.currentStatus = status
        self._updateStatusView()
        if hasattr(self, "sigTMStatusChanged"):
            self.sigTMStatusChanged.emit(status)

        _info = status
        if info is not None:
            _info += " - " + info

        self.informationReceived(_info)

        if status == "acquisition ended":
            # info should be the directory path
            assert info is not None
            assert type(info) is str
            logger.processEnded(
                "Find a valid scan",
                extra={
                    logconfig.DOC_TITLE: self._scheme_title,
                    logconfig.SCAN_ID: info,
                },
            )
            try:
                scan = ScanFactory.create_scan_object(scan_path=info)
            except Exception as e:
                logger.warning(f"Fail to create scan object from {info}. Reason is {e}")
            else:
                self._signalScanReady(scan=scan)

    def _signalScanReady(self, scan):
        assert isinstance(scan, (TomwerScanBase, BlissScan))
        self.lastFoundScans.add(scan)
        self.sigScanReady.emit(scan)  # pylint: disable=E1101

    def mockObservation(self, folder):
        # simple mocking emitting a signal to say that the given folder is valid
        self._setCurrentStatus(status="acquisition ended", info=folder)

    def _updateStatusView(self):
        pass

    def _setMaxAdvancement(self, max):
        """ """
        pass

    def _advance(self, nb):
        """Update the progress bar"""
        pass

    def set_configuration(self, properties):
        for observe_folder_alias in ("observed_folder", "folderObserved"):
            if observe_folder_alias in properties:
                self.setFolderObserved(properties[observe_folder_alias])

    def waitForObservationFinished(self):
        if self.observationThread is not None:
            self.observationThread.waitForObservationFinished()
            self.observationThread.quitEvent.set()

    def getIgnoredFolders(self):
        if self.observationThread is None:
            return []
        else:
            return self.observationThread.observations.ignoredFolders


class DataWatcher(_DataWatcher):
    """For now to avoid multiple inheritance from QObject with the process
    widgets
    we have to define two classes. One only for the QObject inheritance
    """

    sigTMStatusChanged = Signal(str)
    """emit when the status of the scan changed"""
    sigScanReady = Signal(object)
    """emit when scan ready"""

    import threading

    scan_found_event = threading.Event()

    def __init__(self):
        _DataWatcher.__init__(self)
        self.scan = None
        """last found scan"""

    def _signalScanReady(self, scan):
        assert isinstance(scan, TomwerScanBase)
        super()._signalScanReady(scan)
        self.scan_found_event.set()

    def _launchObservation(self):
        """Main function of the widget"""
        if self.isObserving is False:
            return

        # manage data watcher observation
        if self.observationThread is None or not self._observation_thread_running():
            if self._initObservation() is False:
                self._setCurrentStatus("failure")
                logger.info(
                    "failed on observation",
                    extra={logconfig.DOC_TITLE: self._scheme_title},
                )
                return

        self._init_check_finished_scan()

        # starting the observation thread
        self.observationThread.start()



from __future__ import annotations

import weakref
from silx.gui import qt

from tomwer.gui import icons
from silx.gui.plot import PlotWidget


class ShowPlotToolBarAction(qt.QAction):
    """
    QAction opening the ColorBarWidget of the specified plot.
    It was not possible to inherit the one from silx because it inherit from a PlotAction which is not the case of this one (n plots...)
    """

    def __init__(self, plots: tuple[PlotWidget], parent=None):
        icon = icons.getQIcon("show_plot_toolbars")
        text = "show plot toolbars"
        super().__init__(parent=parent, icon=icon, text=text)

        self.setToolTip("Show/Hide plot toolbars")
        self.setCheckable(True)
        self.setChecked(True)

        self._plots = [weakref.ref(plot) for plot in plots]

        self.triggered[bool].connect(self._actionTriggered)

    def _actionTriggered(self, checked=False):
        """Create a cmap dialog and update active image and default colormap."""
        for ref_plot in self._plots:
            plot = ref_plot()
            if plot:
                plot.setToolBarsVisible(checked)



from __future__ import annotations

from silx.gui import qt
from tomwer.gui import icons

try:
    from silx.gui.widgets.OverlayMixIn import OverlayMixIn as _OverlayMixIn
except ImportError:
    from tomwer.third_part.OverlayMixIn import OverlayMixIn as _OverlayMixIn


class ProcessingOverlay(_OverlayMixIn, qt.QLabel):
    """
    Small widget overlay that will display a red dot when some processing are on-going.

    Use case: when loading a frame on the fast axis we cannot clear the plot and add a message else it will
    do some blinking.
    But in some cases loading can take some time (when loading already a full volume in hdf5 for example).
    And we need to notify the user that something is on-going.
    """

    def __init__(self, parent, img_size, *args, **kwargs):
        qt.QLabel.__init__(self, parent=parent)
        _OverlayMixIn.__init__(self, parent=parent, *args, **kwargs)
        self.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignTop)
        self._imageQIcon = icons.getQIcon("red_dot")

        self.setPixmap(self._imageQIcon.pixmap(img_size))



import os
import pickle
import shutil
import tempfile

import numpy

from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.stitching.ZStitchingConfigOW import ZStitchingConfigOW
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.core.volume.hdf5volume import HDF5Volume


class TestScansZStitchingConfigOW(TestCaseQt):
    """
    Test stitching with scans
    """

    def setUp(self):
        super().setUp()
        self._widget = ZStitchingConfigOW()
        self._tmp_dir = tempfile.mkdtemp()
        self._scans = []
        for i_scan in range(3):
            scan = MockNXtomo(
                scan_path=os.path.join(self._tmp_dir, f"scan_{i_scan}.nx"),
                n_ini_proj=20,
                n_proj=20,
                n_alignement_proj=2,
                dim=10,
            ).scan
            self._scans.append(scan)
        self._widget.setConfiguration(
            {
                "inputs": {
                    "input_datasets": [
                        scan.get_identifier().to_str() for scan in self._scans
                    ],
                },
                "stitching": {
                    "type": "z-preproc",
                    "stitching_strategy": "closest",
                    "alignment_axis_1": "center",
                    "alignment_axis_2": "center",
                    "pad_mode": "constant",
                },
                "preproc": {
                    "data_path": "myentry",
                    "location": "/path/to/mynexusfile.nx",
                },
            }
        )
        self.qapp.processEvents()

    def test_get_configuration(self):
        """Make sure the set configuration will be the get configuration"""
        config = self._widget.getConfiguration()
        assert config == {
            "inputs": {
                "input_datasets": list(
                    [scan.get_identifier().to_str() for scan in self._scans]
                ),
                "slices": "",
            },
            "stitching": {
                "stitching_strategy": "cosinus weights",
                "type": "z-preproc",
                "axis_0_pos_px": [0] * len(self._scans),
                "axis_0_params": "overlap_size=;img_reg_method=nabu-fft;window_size=400",
                "axis_1_pos_px": [0] * len(self._scans),
                "axis_1_params": "img_reg_method=None;window_size=400",
                "flipud": False,
                "fliplr": False,
                "avoid_data_duplication": False,
                "rescale_frames": False,
                "rescale_params": "rescale_min_percentile=0;rescale_max_percentile=100",
                "alignment_axis_1": "center",
                "alignment_axis_2": "center",
                "pad_mode": "constant",
            },
            "preproc": {
                "data_path": "myentry",
                "location": "/path/to/mynexusfile.nx",
            },
            "output": {
                "overwrite_results": True,
            },
            "normalization_by_sample": {
                "active": False,
                "margin": 0,
                "method": "median",
                "side": "left",
                "width": 30,
            },
        }
        # test setting some metadata for stitching
        for i_scan, scan in enumerate(
            self._widget.widget._widget._mainWidget.getTomoObjs()
        ):
            scan.stitching_metadata.setPxPos(-i_scan, axis=0)
            scan.stitching_metadata.setPxPos(-i_scan * 10, axis=1)
        self.qapp.processEvents()
        config = self._widget.getConfiguration()
        assert config == {
            "inputs": {
                "input_datasets": list(
                    [scan.get_identifier().to_str() for scan in self._scans]
                ),
                "slices": "",
            },
            "stitching": {
                "type": "z-preproc",
                "stitching_strategy": "cosinus weights",
                "axis_0_pos_px": [0, -1, -2],
                "axis_0_params": "overlap_size=;img_reg_method=nabu-fft;window_size=400",
                "axis_1_pos_px": [0, -10, -20],
                "axis_1_params": "img_reg_method=None;window_size=400",
                "flipud": False,
                "fliplr": False,
                "avoid_data_duplication": False,
                "rescale_frames": False,
                "rescale_params": "rescale_min_percentile=0;rescale_max_percentile=100",
                "alignment_axis_1": "center",
                "alignment_axis_2": "center",
                "pad_mode": "constant",
            },
            "preproc": {
                "data_path": "myentry",
                "location": "/path/to/mynexusfile.nx",
            },
            "output": {
                "overwrite_results": True,
            },
            "normalization_by_sample": {
                "active": False,
                "margin": 0,
                "method": "median",
                "side": "left",
                "width": 30,
            },
        }
        config["stitching"]["axis_1_pos_px"] = [0, -1, -2]
        config["output"]["overwrite_results"] = False
        self._widget.setConfiguration(config)
        assert self._widget.getConfiguration() == config

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        shutil.rmtree(self._tmp_dir)

    def test_serializing(self):
        pickle.dumps(self._widget.getConfiguration())

    def test_literal_dumps(self):
        literal_dumps(self._widget.getConfiguration())


class TestVolumesZStitchingConfigOW(TestCaseQt):
    """
    Test ZStitchingConfigOW with volumes
    """

    def setUp(self):
        super().setUp()
        self._widget = ZStitchingConfigOW()
        self._tmp_dir = tempfile.mkdtemp()
        self._volumes = []
        for i_volumes in range(4):
            volume = HDF5Volume(
                file_path=os.path.join(self._tmp_dir, f"volume_{i_volumes}.hdf5"),
                data_path="entry0000",
                data=numpy.ones(shape=(10, 20, 30)),
                metadata={},
            )
            volume.save()
            self._volumes.append(volume)

        self.output_volume = HDF5Volume(
            file_path=os.path.join(self._tmp_dir, "result.hdf5"),
            data_path="stitched",
        )
        self._widget.setConfiguration(
            {
                "inputs": {
                    "input_datasets": [
                        volume.get_identifier().to_str() for volume in self._volumes
                    ],
                },
                "stitching": {
                    "type": "z-postproc",
                    "stitching_strategy": "closest",
                    "alignment_axis_1": "center",
                    "alignment_axis_2": "center",
                    "pad_mode": "constant",
                },
                "postproc": {
                    "output_volume": self.output_volume.get_identifier().to_str(),
                },
            }
        )
        self.qapp.processEvents()

    def test_get_configuration(self):
        """Make sure the set configuration will be the get configuration"""
        config = self._widget.getConfiguration()
        assert config == {
            "inputs": {
                "input_datasets": list(
                    [scan.get_identifier().to_str() for scan in self._volumes]
                ),
                "slices": "",
            },
            "stitching": {
                "stitching_strategy": "cosinus weights",
                "type": "z-postproc",
                "axis_0_pos_px": [0] * len(self._volumes),
                "axis_0_params": "overlap_size=;img_reg_method=nabu-fft;window_size=400",
                "axis_1_pos_px": [0] * len(self._volumes),
                "axis_1_params": "img_reg_method=None;window_size=400",
                "flipud": False,
                "fliplr": False,
                "avoid_data_duplication": False,
                "rescale_frames": False,
                "rescale_params": "rescale_min_percentile=0;rescale_max_percentile=100",
                "alignment_axis_1": "center",
                "alignment_axis_2": "center",
                "pad_mode": "constant",
            },
            "postproc": {
                "output_volume": self.output_volume.get_identifier().to_str(),
            },
            "output": {
                "overwrite_results": True,
            },
            "normalization_by_sample": {
                "active": False,
                "margin": 0,
                "method": "median",
                "side": "left",
                "width": 30,
            },
        }
        # test setting some metadata for stitching
        for i_scan, scan in enumerate(
            self._widget.widget._widget._mainWidget.getTomoObjs()
        ):
            scan.stitching_metadata.setPxPos(-i_scan, axis=0)
            scan.stitching_metadata.setPxPos(-i_scan * 10, axis=1)
        self.qapp.processEvents()
        config = self._widget.getConfiguration()
        assert config == {
            "inputs": {
                "input_datasets": list(
                    [scan.get_identifier().to_str() for scan in self._volumes]
                ),
                "slices": "",
            },
            "stitching": {
                "stitching_strategy": "cosinus weights",
                "type": "z-postproc",
                "axis_0_pos_px": [0, -1, -2, -3],
                "axis_0_params": "overlap_size=;img_reg_method=nabu-fft;window_size=400",
                "axis_1_pos_px": [0, -10, -20, -30],
                "axis_1_params": "img_reg_method=None;window_size=400",
                "flipud": False,
                "fliplr": False,
                "avoid_data_duplication": False,
                "rescale_frames": False,
                "rescale_params": "rescale_min_percentile=0;rescale_max_percentile=100",
                "alignment_axis_1": "center",
                "alignment_axis_2": "center",
                "pad_mode": "constant",
            },
            "postproc": {
                "output_volume": self.output_volume.get_identifier().to_str(),
            },
            "output": {
                "overwrite_results": True,
            },
            "normalization_by_sample": {
                "active": False,
                "margin": 0,
                "method": "median",
                "side": "left",
                "width": 30,
            },
        }
        # update some field to enhance test
        config["stitching"]["axis_1_pos_px"] = [0, -100, -200, -300]
        config["stitching"]["rescale_frames"] = True
        config["stitching"][
            "rescale_params"
        ] = "rescale_min_percentile=10;rescale_max_percentile=35"
        config["output"]["overwrite_results"] = False
        self._widget.setConfiguration(config)
        assert self._widget.getConfiguration() == config

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        shutil.rmtree(self._tmp_dir)

    def test_serializing(self):
        pickle.dumps(self._widget.getConfiguration())

    def test_literal_dumps(self):
        literal_dumps(self._widget.getConfiguration())



# coding: utf-8
from ewokscore.task import Task as EwoksTask


class _VolumeViewerPlaceHolder(EwoksTask, input_names=("data",)):
    """
    Task to plot a (reconstructed) volume
    """

    def run(self):
        pass



import os
from enum import Enum as _Enum


__LBSRAM_PATH = "/lbsram"

__DEST_PATH = ""

MOCK_LBSRAM = False

MAX_MEM_USED = 80
"Maximal percentage of used memory from which we should skip some processing"

MAKE_OAR_PYST2_PATH = "/data/id19/inhouse/OAR_UTILITIES/pyhst/make_oar_pyhst2"
"""path to the make_oar_pyhst2 path"""

JSON_RPC_HOST = None
"""you specify an host name. If None is provided then will use HOSTNAME"""

JSON_RPC_PORT = 4000

TOMO_BEAMLINES = sorted(
    ("bm05", "bm18", "id11", "id15a", "id16a", "id16b", "id17", "id19")
)


def mock_lsbram(b):
    assert type(b) is bool
    global MOCK_LBSRAM
    MOCK_LBSRAM = b


def isOnLbsram(scan=None):
    if MOCK_LBSRAM:
        return True
    elif scan is None:
        return os.path.isdir(get_lbsram_path())
    else:
        if isinstance(scan, str):
            return os.path.abspath(scan).startswith(__LBSRAM_PATH)
        else:
            return os.path.abspath(scan.path).startswith(__LBSRAM_PATH)


def get_lbsram_path():
    return __LBSRAM_PATH


def get_dest_path():
    return __DEST_PATH


def _set_lbsram_path(path):
    global __LBSRAM_PATH
    __LBSRAM_PATH = path


def _set_dest_path(path):
    global __DEST_PATH
    __DEST_PATH = path


class SlurmSettingsMode(_Enum):
    MANUAL = "manual"
    GENERIC = "generic"
    CAST_VOLUME = "cast_volume"
    SLICE_RECONSTRUCTION = "slice_reconstruction"
    VOLUME_RECONSTRUCTION = "volume_reconstruction"
    STITCHING = "stitching"

    @staticmethod
    def get_settings_class(mode):
        assert isinstance(mode, SlurmSettingsMode)
        if mode is SlurmSettingsMode.MANUAL:
            return None
        elif mode is SlurmSettingsMode.GENERIC:
            return SlurmSettings
        elif mode is SlurmSettingsMode.CAST_VOLUME:
            return DefaultSlurmSettingsCastVolume
        elif mode is SlurmSettingsMode.SLICE_RECONSTRUCTION:
            return DefaultSlurmSettingsSliceReconstruction
        elif mode is SlurmSettingsMode.VOLUME_RECONSTRUCTION:
            return DefaultSlurmSettingsVolumeReconstruction
        elif mode is SlurmSettingsMode.STITCHING:
            return DefaultSlurmSettingsStitching
        else:
            raise ValueError(f"{mode} not handled")


class SlurmSettings:
    # Default slurm cluster configuration

    N_CORES_PER_TASK = 16
    """Number of CPU per worker"""

    N_TASKS = 1
    """Number of worker"""

    N_JOBS = 1
    """on how many job we want to split the EwoksTask"""

    MEMORY_PER_WORKER = 128  # memory in GB
    """Amount of memory per worker"""

    PARTITION = "gpu"
    """Queue / partition to use"""

    DEFAULT_WALLTIME = "01:00:00"  # None if the default wall time

    N_GPUS_PER_WORKER = 1
    """number of gpu per worker"""

    PYTHON_VENV = None
    """Python executable to take. Useful if compute nodes have a different environment from the front end.
    """

    PROJECT_NAME = "tomwer_{scan}_-_{process}_-_{info}"
    """Slurm cluster project name. `scan`, `process` and `info` will be format.
    """

    MODULES_TO_LOAD = ("tomotools/stable",)

    SBATCH_EXTRA_PARAMS = {
        "export": "ALL",  # value to provide to sbatch --export={}
    }


class DefaultSlurmSettingsCastVolume(SlurmSettings):
    """
    default proposed configuration for casting a volume remotely
    """

    N_JOBS = 1

    N_GPUS_PER_WORKER = 0

    MEMORY_PER_WORKER = 128  # memory in GB

    PARTITION = "nice-long"


class DefaultSlurmSettingsSliceReconstruction(SlurmSettings):
    """
    default proposed configuration for reconstructing a single slice remotely
    """

    N_JOBS = 1

    N_GPUS_PER_WORKER = 1

    MEMORY_PER_WORKER = 220  # memory in GB


class DefaultSlurmSettingsVolumeReconstruction(DefaultSlurmSettingsSliceReconstruction):
    pass


class DefaultSlurmSettingsStitching(SlurmSettings):
    """
    default proposed configuration for stitching
    """

    N_CORES_PER_TASK = 1

    N_JOBS = 16

    N_GPUS_PER_WORKER = 0

    PARTITION = "nice-long"



from __future__ import annotations

from silx.gui import qt
from tomwer.gui import icons

try:
    from silx.gui.widgets.OverlayMixIn import OverlayMixIn as _OverlayMixIn
except ImportError:
    from tomwer.third_part.OverlayMixIn import OverlayMixIn as _OverlayMixIn


class CoordinateSystemOverlay(_OverlayMixIn, qt.QLabel):
    "Widget to be display an image on one of the four widget corner"

    def __init__(self, parent, img, img_size, *args, **kwargs):
        qt.QLabel.__init__(self, parent=parent)
        _OverlayMixIn.__init__(self, parent=parent, *args, **kwargs)
        self._imageQIcon = icons.getQIcon(img)

        self.setPixmap(self._imageQIcon.pixmap(img_size))



from orangewidget import gui
from orangewidget.settings import Setting

from silx.gui import qt
from ewoksorange.bindings.owwidgets import OWEwoksWidgetNoThread

import tomwer.core.process.control.emailnotifier
from tomwer.core.utils.dictutils import concatenate_dict

try:
    from ewoksnotify.gui.email import EmailWidget
except ImportError:
    has_ewoksnotify = False
else:
    has_ewoksnotify = True


class EmailOW(
    OWEwoksWidgetNoThread,
    ewokstaskclass=tomwer.core.process.control.emailnotifier.TomoEmailTask,
):
    """
    This widget will browse a folder and sub folder to find valid tomo scan project.
    Contrary to the scan watcher it will parse all folder / sub folders then stop.
    """

    name = "email notifier"
    id = "orangecontrib.widgets.tomwer.control.EmailOW.EmailOW"
    description = (
        "This widget will send an email to receivers when the input is provided. \n"
    )
    icon = "icons/email.svg"
    priority = 146
    keywords = [
        "tomography",
        "tomwer",
        "tomo_obj",
        "email",
        "notifier",
        "notification",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _ewoks_default_inputs = Setting({"configuration": {}, "tomo_obj": None})

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not has_ewoksnotify:
            raise ImportError(
                "ewoksnotify not install but required. Please run 'pip install ewoksnotify[full]'"
            )
        self._widget = EmailWidget(parent=self)

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)

        # load settings
        self._widget.setConfiguration(self._ewoks_default_inputs)

        # connect signal / slot
        self._widget.sigChanged.connect(self._updateSettings)

    def _updateSettings(self):
        self._ewoks_default_inputs = {
            "configuration": self.getConfiguration(),
            "tomo_obj": None,
        }

    # expose some API
    def getConfiguration(self) -> dict:
        return self._widget.getConfiguration()

    def setConfiguration(self, config: dict):
        self._widget.setConfiguration(config.get("configuration", {}))

    def get_task_inputs(self):
        return concatenate_dict(
            super().get_task_inputs(),
            {"configuration": self.getConfiguration()},
        )

    def sizeHint(self):
        return qt.QSize(500, 200)

    def _execute_ewoks_task(self, *args, **kwargs):
        arguments = self._get_task_arguments()
        if arguments.get("inputs", {}).get("tomo_obj", None) is not None:
            super()._execute_ewoks_task(*args, **kwargs)



# coding: utf-8
from __future__ import annotations


import shutil
import tempfile
import unittest

import h5py
import numpy
from tomoscan.normalization import Method

from tomwer.core.process.reconstruction.normalization import normalization, params
from tomwer.core.utils.scanutils import MockNXtomo


class TestNormalization(unittest.TestCase):
    """
    Test the normalization process
    """

    def setUp(self) -> None:
        self.tempdir = tempfile.mkdtemp()
        dim = 100
        self.scan = MockNXtomo(
            scan_path=self.tempdir,
            n_proj=2,
            n_ini_proj=2,
            scan_range=180,
            dim=dim,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
        ).scan

    def tearDown(self) -> None:
        shutil.rmtree(self.tempdir)

    def testManualROI(self):
        # step1: rewrite the detector data to simplify result check
        with h5py.File(self.scan.master_file, mode="a") as h5f:
            dataset = h5f["/entry/instrument/detector/data"]
            assert dataset.shape == (2, 100, 100)  # pylint: disable=E1101
            del h5f["/entry/instrument/detector/data"]
            h5f["/entry/instrument/detector/data"] = numpy.arange(
                100 * 100 * 2
            ).reshape(2, 100, 100)

        process_params = normalization.SinoNormalizationParams()
        process_params.method = Method.SUBTRACTION
        process_params.source = params._ValueSource.MANUAL_ROI
        expected_results = {
            "mean": numpy.array([800.5, 10800.5]),
            "median": numpy.array([800.5, 10800.5]),
        }

        for item in params._ValueCalculationFct:
            calc_fct = item.value
            with self.subTest(calc_fct=calc_fct):
                process_params.extra_infos = {
                    "start_x": 0,
                    "end_x": 2,
                    "start_y": 8,
                    "end_y": 9,
                    "calc_fct": calc_fct,
                    "calc_area": "volume",
                }
                process = normalization.SinoNormalizationTask(
                    inputs={
                        "data": self.scan,
                        "configuration": process_params,
                        "serialize_output_data": False,
                    }
                )
                process.run()
                res = self.scan.intensity_normalization.get_extra_infos().get("value")
                if isinstance(res, numpy.ndarray):
                    numpy.testing.assert_array_equal(res, expected_results[calc_fct])
                else:
                    numpy.testing.assert_array_equal(res, expected_results[calc_fct])



import gc
import logging

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.other.PythonScriptOW import OWPythonScript
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.volume.hdf5volume import HDF5Volume

logging.disable(logging.INFO)


class TestEDFDarkRefWidget(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)
        self.widget = OWPythonScript()
        self._scan = NXtomoScan("my_file.hdf5", entry="entry0000")
        self._volume = HDF5Volume(file_path="my_volume.hdf5", data_path="volume")

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        gc.collect()

    def test_handle_data(self):
        self.widget.handle_input(self._scan, (1,), "data")

    def test_handle_volume(self):
        self.widget.handle_input(self._volume, (1,), "volume")
